# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya, soya.widget as widget, soya.ray as ray
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
from soya import Point, Vector
import slune.character as slune_character, slune.globdef as globdef
import slune.controler


def init_fight_game(level):
  pass

def start_fight_game(level):
  CheckEveryoneIsKilled(level)
  
def end_fight_game(level):
  pass

class CheckEveryoneIsKilled(soya.Volume):
  def begin_round(self):
    from py2play.player import CURRENT_PLAYER
    
    if (len(self.parent.characters) == 1) and (self.parent.characters[0] is CURRENT_PLAYER.character):
      from py2play.idler import IDLER
      
      IDLER.level_completed(CURRENT_PLAYER.character, 1)
      
      self.parent.remove(self)
      

STRATEGY_RUN      = 0
STRATEGY_FLEE     = 1
STRATEGY_BACKSTAB = 2
STRATEGY_DODGE    = 3
STRATEGY_FRONTRUN = 4
STRATEGY_SIDERUN  = 5
STRATEGY_WANDER   = 6

class FightChromosom:
  def __init__(self, max_speed, will, turbosity, autododge, run, dodge, flee, frontrun, siderun, wander):
    self.max_speed   = max_speed
    self.will        = will
    self.turbosity   = turbosity
    self.autododge   = autododge
    self.run         = run
    self.dodge       = dodge
    self.flee        = flee
    self.frontrun    = frontrun
    self.siderun     = siderun
    self.wander      = wander
    
  def mutate(self, random):
    for attr, value in self.__dict__.iteritems():
      if attr != "max_speed":
        if random.random() < 0.4: 
          setattr(self, attr, value * (0.5 + random.random()))
          
CHROMOSOMS = {
  
  ("tux", 0):    FightChromosom(0.65, 100.0, -7.0, 0.0, 0.8, 1.0, 0.1, 1.7, 1.0, 1.3),
  ("tux", 1):    FightChromosom(0.8,  100.0, -1.0, 0.0, 1.0, 1.1, 0.6, 1.1, 1.3, 1.0),
  ("tux", 2):    FightChromosom(0.9,  100.0,  0.5, 2.5, 1.0, 1.5, 1.1, 1.0, 1.5, 0.5),
  
  ("gnu", 0):    FightChromosom(0.6,  150.0, -2.0, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5),
  ("gnu", 1):    FightChromosom(0.75, 150.0,  0.7, 0.0, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1),
  ("gnu", 2):    FightChromosom(0.9,  150.0,  1.0, 1.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5),
  
  ("shark", 0):  FightChromosom(0.6,   80.0, -9.0, 0.0, 0.9, 1.0, 0.2, 1.6, 0.5, 1.5),
  ("shark", 1):  FightChromosom(0.7,   80.0, -4.0, 1.0, 1.0, 1.4, 0.5, 1.0, 0.5, 1.2),
  ("shark", 2):  FightChromosom(0.9,   70.0,  0.0, 1.5, 1.0, 1.5, 0.6, 0.2, 0.5, 0.9),

  ("python", 0): FightChromosom(0.6,  100.0,  0.0, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5),
  ("python", 1): FightChromosom(0.8,  100.0,  1.3, 0.5, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1),
  ("python", 2): FightChromosom(0.9,  100.0,  2.0, 2.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5),
  
  
#   ("tux", 0): [0.23161940934461928, 0.0, 0.0, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 2.0, 0.0, 0.8, 1.0, 0.1, 1.7, 1.0, 1.3, 0.65, 20.0],
#   ("tux", 1): [0.23161940934461928, 0.0, 0.0, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 3.0, 0.0, 1.0, 1.1, 0.6, 1.1, 1.3, 1.0, 0.8 , 10.0],
#   ("tux", 2): [0.23161940934461928, 0.0, 0.0, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 4.0, 2.5, 1.0, 1.5, 1.1, 1.0, 1.5, 0.5, 0.9 ,  1.0],
  
#   ("gnu", 0): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 2.0, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5, 0.6 , 30.0],
#   ("gnu", 1): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.0, 0.0, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1, 0.75, 20.0],
#   ("gnu", 2): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.5, 1.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5, 0.9 , 3.0],
  
#   ("shark", 0): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 2.0, 0.0, 0.9, 1.0, 0.2, 1.6, 0.5, 1.5, 0.6, 20.0],
#   ("shark", 1): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.0, 1.0, 1.0, 1.4, 0.5, 1.0, 0.5, 1.2, 0.7, 12.0],
#   ("shark", 2): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.5, 1.5, 1.0, 1.5, 0.6, 0.2, 0.5, 0.9, 0.9,  2.0],

# For mission levels
#  ("killer", 0): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 2.0, 0.0, 1.2, 0.0, 0.0, 0.0, 0.6, 0.5, 0.85, 1.0],
#  ("killer", 1): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.0, 0.0, 1.8, 0.0, 0.0, 0.0, 0.7, 0.2, 0.92,  1.0],
#  ("killer", 2): [0.27955060118474129, 0.0, 0.0, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 3.5, 0.0, 2.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.05,  1.0],
  
  ("killer", 0):    FightChromosom(0.6,  150.0, 0.2, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5),
  ("killer", 1):    FightChromosom(0.75, 150.0, 0.7, 0.0, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1),
  ("killer", 2):    FightChromosom(0.9,  150.0, 1.0, 1.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5),
  
  ("killer_flight", 0): FightChromosom(0.6,  150.0, -10.0, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5),
  ("killer_flight", 1): FightChromosom(0.75, 150.0, -10.0, 0.0, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1),
  ("killer_flight", 2): FightChromosom(0.9,  150.0, -10.0, 1.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5),
  
  ("alien_boss", 0):    FightChromosom(0.9, 150.0, 1.5, 1.5, 2.0, 0.5, 0.4, 0.5, 1.0, 0.5),
  ("alien_boss", 1):    FightChromosom(0.9, 150.0, 1.0, 0.0, 1.8, 0.6, 0.3, 1.2, 0.7, 1.1),
  ("alien_boss", 2):    FightChromosom(0.9, 150.0, 0.7, 0.0, 1.2, 1.0, 0.1, 2.0, 0.6, 1.5),
  }

class Fighter(slune_character.NonPlayerCharacter):
  def __init__(self, parent = None, chromosom = None):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    
    self.chromosom = chromosom
    self.clan = 2
    
  def choose_target(self):
    best      = None
    best_dist = 10000.0
    for level in self.get_root():
      if isinstance(level, py2p_level.Level):
        for character in level.characters:
          if (not character is self) and (character.clan > 0) and (character.clan != self.clan):
            dist = character.distance_to(self)
            if dist < best_dist:
              best      = character
              best_dist = dist
    return best
  
  def choose_strategy(self, target):
    dist = self.distance_to(target)
    c = self.chromosom
    return max(
      (self.life              + self.random.random() * c.run,      STRATEGY_RUN),
      (self.life - 0.1 * dist + self.random.random() * c.dodge,    STRATEGY_DODGE),
      (1.0 - 2.0 * self.life  + self.random.random() * c.flee,     STRATEGY_FLEE),
      (self.life              + self.random.random() * c.frontrun, STRATEGY_FRONTRUN),
      (self.life              + self.random.random() * c.siderun,  STRATEGY_SIDERUN),
      (                         self.random.random() * c.wander,   STRATEGY_WANDER),
      )[1]
  
  def init(self):
    if not self.perso_name:
      persos = list(globdef.CHARACTERS)
      self.random.shuffle(persos)
      for perso in persos:
        perso = perso.lower()
        for c in self.level:
          if (hasattr(c, "perso_name")) and (c.perso_name == perso): break
        else:
          self.perso_name = perso
          break
      else: self.perso_name = persos[0].lower()
    self.set_perso(self.perso_name)
    self.set_vehicle(self.random.choice(globdef.VEHICLES.values()))
    
  def IAControler(self):
    c = self.chromosom = self.chromosom or CHROMOSOMS[self.perso_name, self.level.difficulty]
    c.mutate(self.random)
    self.max_speed = c.max_speed
    
    target = None
    while 1:
      target   = self.choose_target() or target
      strategy = self.choose_strategy(target)
      
      if   strategy == STRATEGY_RUN:
        yield slune.controler.AttackRun(self, target, c.will, c.autododge, c.turbosity + 2.0 * self.life)
        
      elif strategy == STRATEGY_FRONTRUN:
        yield slune.controler.AttackFrontRun(self, target, c.will, c.autododge, c.turbosity + 2.0 * self.life)
        
      elif strategy == STRATEGY_SIDERUN:
        yield slune.controler.AttackSideRun(self, target, c.will, c.autododge, c.turbosity + 2.0 * self.life)
        
      elif strategy == STRATEGY_FLEE:
        yield slune.controler.Flee(self, target, c.will)
        
      elif strategy == STRATEGY_DODGE:
        yield slune.controler.Feinte(self, target, c.will, c.autododge + 3.0)
        
      elif strategy == STRATEGY_WANDER:
        yield slune.controler.Wander(self, c.will)

        
class AlienBoss(Fighter, slune_character.Competitor):
  def __init__(self, parent = None):
    from py2play.player import CURRENT_PLAYER
    Fighter.__init__(self, parent, CHROMOSOMS["alien_boss", CURRENT_PLAYER.level.difficulty])


class Mission6Killer(slune_character.NonPlayerCharacter):
  def __init__(self, parent = None, chromosom = None):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    
    self.chromosom = chromosom
    self.clan = 2
    
  def choose_target(self):
    best      = None
    best_dist = 10000.0
    for level in self.get_root():
      if isinstance(level, py2p_level.Level):
        for character in level.characters:
          if (not character is self) and (character.clan > 0) and (character.clan != self.clan):
            dist = character.distance_to(self)
            if dist < best_dist:
              best      = character
              best_dist = dist
    return best
  
  def init(self):
    self.set_perso(self.perso_name)
    
  def IAControler(self):
    c = self.chromosom = self.chromosom or CHROMOSOMS[self.perso_name, self.level.difficulty]
    self.max_speed = c.max_speed
    
    from py2play.idler import IDLER
    
    while 1:
      target = self.choose_target() or target
      if not hasattr(target, "nb_avertissements"):
        target.disabled          = 0
        target.nb_avertissements = {0:4, 1:2, 2:0}[target.level.difficulty]
        
      for i in range(100):
        fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
        
        if   target.disabled: target.disabled -= 1
        elif (target.y < 1.4) and (fz < 0.0) and (self.distance_to(target) < 3.0):
          IDLER.message(_("__scenar6-1__"))
          if target.nb_avertissements:
            target.disabled = 100
            target.nb_avertissements -= 1
          else: # You loose...
            from videosequence import GnuSpeech
            GnuSpeech((_("__scenar6-2__"), _("__gameover__")), 1).start(tux = target)
            
            while 1: yield slune_character.Action()
            
        yield slune.controler.runto(self, target, canjump = 0)
        

class Mission7Pusher(slune_character.NonPlayerCharacter):
  def __init__(self, parent = None, chromosom = None):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    
    self.chromosom = chromosom
    
    self.clan = 2
    self.check_ground_vector = Vector(self, 0.0, -2.0, -2.0)
    self.jumping = 0
    self.current_action = slune_character.Action()
    
  def choose_target(self):
    best      = None
    best_dist = 10000.0
    for level in self.get_root():
      if isinstance(level, py2p_level.Level):
        for character in level.characters:
          if (not character is self) and (character.clan > 0) and (character.clan != self.clan):
            dist = character.distance_to(self)
            if dist < best_dist:
              best      = character
              best_dist = dist
    return best
  
  def init(self): self.set_perso(self.perso_name)
  
  def IAControler(self):
    c = self.chromosom = self.chromosom or CHROMOSOMS[self.perso_name, self.level.difficulty]
    self.max_speed = c.max_speed
    
    from py2play.idler import IDLER
    
    for i in range([10000, 10, 6][self.parent.difficulty]):
      target = self.choose_target() or target
      yield slune.controler.PushToKill(self, target, c.turbosity + 2.0 * self.life)
      
    while 1: # Turn berserk
      target = self.choose_target() or target
      yield slune.controler.AttackRun(self, target, turbosity = c.turbosity + 2.0 * self.life)
      




# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import soya, slune.globdef as globdef
import py2play.idler as idler
from py2play.player import ActivePlayer
import py2play.player, py2play.level
from slune.character import PlayerCharacter
from slune.controler import *
from slune.sound import end_music


def start_new_game(level_name):
  #soya.init(width = globdef.SCREEN_WIDTH, height = globdef.SCREEN_HEIGHT)
  #soya.cursor_set_visible(0)
  #if globdef.FULLSCREEN: soya.toggle_fullscreen()
  
  import slune.level
  idler = slune.level.Idler()
  idler.show_fps()
  idler.camera.fov  = 60.0
  idler.camera.back = 50.0
  
  # HACK ! Needed because CURRENT_PLAYER == None means that we are in "editor mode" and not in "game mode" (see slune.level.Level.__setstate__).
  py2play.player.CURRENT_PLAYER = 1
  
  level  = py2play.level.CREATE(level_name)
  player = create_player(globdef.NAME, globdef.PORT, level)
  
  if level.filename.startswith("level-mission"):
    player.character.set_perso("tux")
    
  level.set_active(1)
  level.filename = None
  
  #idler.camera.add_traveling(soya.ThirdPersonTraveling(player.character.camera_target))
  t = soya.ThirdPersonTraveling(player.character.camera_target)
  t.offset_y = 0.9
  idler.camera.add_traveling(t)
  
  idler.init_game()
  
  
  #idler.camera.add_traveling(soya.FixTraveling(None, Vector(None, 0.0, 0.0, 1.0)))
  #idler.camera.add_traveling(soya.FixTraveling(None, Vector(None, 1.0, 0.0, 0.0)))
  #idler.camera.add_traveling(soya.FixTraveling(None, Vector(None, -1.0, 0.0, 0.0)))
  #idler.camera.set_xyz(290.0, 22.0, 130.0)
  
#   for e in level.recursive():
#     if isinstance(e, soya.Light):
#       e.look_at(soya.Vector(None, 0.0, -1.0, 1.0))
#   w = soya.World()
#   f = soya.Face(w, [
#     soya.Vertex(w, -10.0, 0.0, -10.0),
#     soya.Vertex(w,  10.0, 0.0, -10.0),
#     soya.Vertex(w,  10.0, 0.0,  10.0),
#     soya.Vertex(w, -10.0, 0.0,  10.0),
#     ])
#   f.double_sided = 1
#   w.shapifier = soya.SimpleShapifier()
#   w.shapifier.shadow = 1
#   v = soya.Volume(level, w.shapify())
#   v.set_xyz(290.0, 30.0, 120.0)
  
  r = idler.idle()
  
  #soya.cursor_set_visible(1)
  #if globdef.FULLSCREEN: soya.toggle_fullscreen()
  
  
  #import gc, weakref, soya.soya3d as soya3d, memtrack
  #idler_ = weakref.ref(idler)
  #level_ = weakref.ref(level)
  #player_= weakref.ref(player)
  
  # Del these ref in order to clean_mem() to be able to work.
  idler = None
  level = None
  player = None
  clean_mem()
  
  end_music()
  
#   if idler_():
#     print "Idler :"
#     for i in gc.get_referrers(idler_()):
#       print
#       print repr(i)
#       print
      
#       #if isinstance(i, dict):
#       #  for k in i.keys():
#       #    if i[k] is idler_(): i[k] = None
#       #if hasattr(i, "parent"):
#       #  i.parent.remove(i)
      
#   if level_():
#     print "Level :"
#     refs = gc.get_referrers(level_())
#     #refs = filter(lambda i: not isinstance(i, soya3d.Volume), refs)
#     #world_ = weakref.ref(filter(lambda i: isinstance(i, soya3d.World), refs)[0])
#     #refs = filter(lambda i: not isinstance(i, soya3d.World), refs)
#     refs = filter(lambda i: not isinstance(i, dict), refs)
    
#     for i in refs:
#       print
#       print repr(i)
#       print
      
#       if isinstance(i, dict):
#         for k in i.keys():
#           if i[k] is level_(): i[k] = None
#       if hasattr(i, "parent"):
#         i.parent.remove(i)
#     i    = None
#     refs = None
#   print
#   print
  
#  gc.collect()
#  gc.collect()
#  gc.collect()
#  gc.collect()
#  gc.collect()
  
#   if idler_():
#     print "idler :", gc.get_referrers(idler_())
#     print len(gc.get_referrers(idler_()))
#     print gc.garbage, idler_()
    
#   print
#  if level_():
#     print "level :"
#     for i in gc.get_referrers(level_()): print i
#     i = None
#     print len(gc.get_referrers(level_()))
#     print level_()

    
#    print
#    memtrack.dump(*memtrack.track(idler_()))
#    #memtrack.dump(*memtrack.reverse_track(idler_()))
    
#   #f = gc.get_referrers(idler_())[1]
#   #print "f", f
#   #print dir(f)
#   #print f.f_locals
    
#   from slune.character import HACK
#   memtrack.dump(*memtrack.reverse_track(HACK()))
  
  
#  print "garbage", gc.garbage
  
  return r

def join_network_game():
  #soya.init(width = globdef.SCREEN_WIDTH, height = globdef.SCREEN_HEIGHT)
  #soya.cursor_set_visible(0)
  #if globdef.FULLSCREEN: soya.toggle_fullscreen()
  
  import slune.level, socket
  idler = slune.level.Idler()
  idler.show_fps()
  idler.camera.fov  = 60.0
  idler.camera.back = 50.0
  
  try:
    player = parrain_by_player(globdef.NAME + "_2", globdef.PORT, globdef.PARRAIN_HOST, globdef.PARRAIN_PORT)
  except socket.error, ex:
    msg = _("Cannot join multiplayer game at %s:%s %s!") % (globdef.PARRAIN_HOST, globdef.PARRAIN_PORT, ex)
    try:
      print "* Slune * " + msg.encode("latin")
    except: pass
    
    if globdef.GUI:
      import tkMessageBox
      tkMessageBox.showerror(_("Network error!"), msg)
      
    return
    
  level = player.level
  
  level.set_active(1)
  level.filename = None
  
  idler.camera.add_traveling(soya.ThirdPersonTraveling(player.character.camera_target))
  idler.init_game()
  
  r = idler.idle()
  
  #soya.cursor_set_visible(1)
  #if globdef.FULLSCREEN: soya.toggle_fullscreen()
  
  end_music()
  return r
  

def create_player(name, port, level):
  player = ActivePlayer(level, name, "", port)
  
  
  character = PlayerCharacter(player)
  player.controler = KeyboardMouseControler(character)
  character.set_level(level, 1)
  
  
  idler.IDLER.lifebar.character = character
  
  return player

def parrain_by_player(name, port, parrain_host, parrain_port = 36079):
  player = ActivePlayer(None, name, "", port)
  player.parrained_by(parrain_host, parrain_port)
  
  character = PlayerCharacter(player)
  player.controler = KeyboardMouseControler(character)
  character.set_level(player.level, 1)
  
  idler.IDLER.lifebar.character = character
  
  return player

def clean_mem():
  import slune.character as slune_character
  
  idler.IDLER = None
  py2play.player.CURRENT_PLAYER = None
  
  slune_character._P  = soya.Point()
  slune_character._P2 = soya.Point()
  slune_character._P3 = soya.Point()
  slune_character._V  = soya.Vector()
  slune_character._V2 = soya.Vector()
  slune_character.explosions *= 0
  
  soya.set_root_widget(None)

  sound.clean_mem()
  
  import gc; gc.collect() # Perform GC after, rather than while, playing !!!



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import math

import soya, soya.sdlconst as soyasdlconst
import py2play.action as action

from slune.character import *
from slune.level     import Level, LifeBonus, FlameThrowerBonus

def Debug(character):
  yield Action(ACTION_DEBUG)
  yield Action(ACTION_DEBUG)
  yield Action(ACTION_DEBUG)
  yield Action(ACTION_DEBUG)
  
def Wall_climber(character):
  yield Action(ACTION_JUMP, character.default_mouse_x(), -1.0)
  
  while character.parent.transform_vector(0.0, 1.0, 0.0, character)[1] > 0.0:
    yield Action(ACTION_ROTUP, character.default_mouse_x(), -1.0)
    
  i = 0
  for i in range(10):
    yield Action(ACTION_ROTDOWN, character.default_mouse_x(), -1.0)
    
  while (not character.on_ground) and (i < 25.0 * 7.0 / character.rotup_speed):
    i += 1
    yield Action(ACTION_ROTDOWN, character.default_mouse_x(), -1.0)
    
def Dodge(character):
  yield Action(ACTION_JUMP, 0.0, 0.0)
  
  for j in range(30):
    yield Action(ACTION_ROTUP, 0.0, 0.0)
    
  for j in range(30):
    yield Action(ACTION_WAIT, 0.0, 0.0)
    
def nearest_character(character):
  best      = None
  best_dist = 10000.0
  for level in character.get_root():
    if isinstance(level, Level):
      for charac in level.characters:
        if not charac is character:
          dist = character.distance_to(charac)
          if dist < best_dist:
            best      = charac
            best_dist = dist
            
  return best
  
GOTO_X_2_TURN         = 0.21412933552126162
GOTO_X_2_SPEED        = 0.49762284312684019
GOTO_Y_2_SPEED        = 0.49762284312684019
GOTO_Z_2_SPEED        = 1.6085952599320903
GOTO_NOGROUND_2_SPEED = -18.661345433135967
GOTO_Y_2_JUMP         = 0.38818320682881224
GOTO_Z_2_JUMP         = 0.059585204340943942
GOTO_NOGROUND_2_JUMP  = 13.608682920401893

def check_vertical_angle(self):
  if self.on_ground and not self.gripped:
    if self.vertical_angle < -45.0:
      self.vertical_angle = 0.0
      return ACTION_ROTDOWN
    
  return ACTION_WAIT

def goto(self, target, coordsyst = None, x = None, y = None, z = None, dodge = 0.0, canjump = 1, turbosity = -100.0, turbo_period = -1.0):
  if coordsyst: fx, fy, fz = self.transform_point(x, y, z, coordsyst)
  else:         fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
  
  if dodge and (fz > 0.0) and (self.distance_to(target) < dodge): return Dodge(self)
  
  if self.jumping:
    if self.on_ground: self.jumping = 0
    no_ground = 0 # It is normal that there is no ground !
  elif self.context:
    no_ground = (not self.on_ground) and (not self.context.raypick_b(self, self.check_ground_vector, 6.0, 3))
  else: no_ground = not self.on_ground

  if self.gripped: # Avoid jumping while gripped
    if canjump and ((GOTO_Y_2_JUMP * fy + GOTO_Z_2_JUMP * fz + GOTO_NOGROUND_2_JUMP * no_ground > 8.0) or (self.speed.length() < 0.001)):
      self.jumping = 1
      action = ACTION_JUMP
    else: action = check_vertical_angle(self)
  else:
    if   canjump and ((fy > 3.5) and (abs(fx) < 2.0) and (-10.0 < fz < 0.0)):
      self.jumping = 1
      return Wall_climber(self)
    elif canjump and ((GOTO_Y_2_JUMP * fy + GOTO_Z_2_JUMP * fz + GOTO_NOGROUND_2_JUMP * no_ground > 2.0) or ((self.speed.length() < 0.001) and (self.current_action.mouse_y < 1.0))):
      self.jumping = 1
      action = ACTION_JUMP
    else: action = check_vertical_angle(self)

  if not self.flame_thrower:
    if (turbosity > 0.0) and (action == ACTION_WAIT) and not self.gripped:
      if turbo_period == -1.0:
        if turbosity * (-fz - abs(fx) - abs(fy)) > 45.0: action = ACTION_TURBO
      else:
        if fz < 0.0:
          if turbosity * (fz - abs(fx) - abs(fy)) > -10.0: action = ACTION_TURBO
      
      
  return Action(
    action,
    (5.0 * self.current_action.mouse_x + min(1.0, max(-1.0, GOTO_X_2_TURN * fx))) / 6.0,
    (5.0 * self.current_action.mouse_y + min(1.0, max(-self.max_speed, 1.0 + GOTO_X_2_SPEED * abs(fx) + GOTO_Y_2_SPEED * fy + GOTO_Z_2_SPEED * fz + GOTO_NOGROUND_2_SPEED * no_ground))) / 6.0,
    )
  
runto = goto
  
def Goto(self, target, min_distance = 0.7):
  while self.distance_to(target) > min_distance:
    yield goto(self, target)
    
Runto = Goto

def CheckBonus(self, aggressivity = 0.5):
  for bonus in self.level.bonuss:
    if bonus.visible and (self.distance_to(bonus) < 25.0):
      if   isinstance(bonus, LifeBonus) and (self.life < 0.6):
        for i in range(250):
          yield goto(self, bonus)
          if (bonus.visible == 0) or (self.distance_to(bonus) < getattr(bonus, "radius", 0.3)): break
      elif isinstance(bonus, FlameThrowerBonus) and (self.random.random() < aggressivity):
        for i in range(250):
          yield goto(self, bonus)
          if (bonus.visible == 0) or (self.distance_to(bonus) < getattr(bonus, "radius", 0.3)): break
          
def AttackRun(self, target, duration = 100, dodge = 0.0, turbosity = -100.0):
  for i in range(int(duration)):
    a =  runto(self, target, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
    if self.flame_thrower and isinstance(a, Action):
      fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
      if (-15.0 < fz < -0.2) and (-0.5 < fy / fz < 0.5) and (-0.5 < fx / fz < 0.5):
        if a.action == ACTION_WAIT: a.action = ACTION_TURBO
    yield a
    
def AttackFrontRun(self, target, duration = 100, dodge = 0.0, turbosity = -100.0):
  for i in range(int(duration)):
    gx, gy, gz = target.transform_point(self.x, self.y, self.z, self.parent)
    if (gz <= 0.0) or self.distance_to(target) > 10.0: a = runto(self, target, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
    else:
      if gx < 0: a = runto(self, target, target, -3.0, 0.0, -7.0, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
      else:      a = runto(self, target, target,  3.0, 0.0, -7.0, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
    if self.flame_thrower and isinstance(a, Action):
      fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
      if (-15.0 < fz < -0.2) and (-0.5 < fy / fz < 0.5) and (-0.5 < fx / fz < 0.5):
        if a.action == ACTION_WAIT: a.action = ACTION_TURBO
    yield a
    
def AttackSideRun(self, target, duration = 100, dodge = 0.0, turbosity = -100.0):
  for i in range(int(duration)):
    gx, gy, gz = target.transform_point(self.x, self.y, self.z, self.parent)
    if (gz <= 0.0) or self.distance_to(target) > 10.0: a = runto(self, target, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
    else:
      if gx < 0: a = runto(self, target, target, -7.0, 0.0, 0.0, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
      else:      a = runto(self, target, target,  7.0, 0.0, 0.0, dodge = dodge, turbosity = turbosity, turbo_period = 1.0)
    if self.flame_thrower and isinstance(a, Action):
      fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
      if (-15.0 < fz < -0.2) and (-0.5 < fy / fz < 0.5) and (-0.5 < fx / fz < 0.5):
        if a.action == ACTION_WAIT: a.action = ACTION_TURBO
    yield a
      
def Flee(self, target, duration = 100):
  no_ground = 0
  
  for i in range(int(duration)):
    fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
    dist       = self.distance_to(target)
    
    if self.jumping:
      if not no_ground: self.jumping = 0
      no_ground = 0 # It is normal that there is no ground !
    else:
      if self.context: no_ground = not self.context.raypick_b(self, self.check_ground_vector, 4.0, 3)
      
    if self.speed.length() * 10.0 + dist < 3.0:
      h = self.random.random()
      if   h < 0.2: break # Try another strategy
      elif h < 0.4: yield Dodge(self)
    
    yield Action(
      ACTION_WAIT,
      (5.0 * self.current_action.mouse_x + min(1.0, max(-1.0, -0.27955060118474129 * fx))) / 6.0,
      (5.0 * self.current_action.mouse_y + min(1.0, max(-self.max_speed, -0.94713263528551006 * abs(fx) + -0.0076398049940541654 * fy + -0.46242007806615826 * fz + 22.758284696585228 * no_ground))) / 6.0,
      )

def Feinte(self, target, duration = 100, dodge = 1.0):
  for i in range(int(duration)):
    if (self.distance_to(target) < dodge):
      yield Dodge(self)
      break # Change strategy !
    
    yield Action(ACTION_WAIT, 0.5, 0.1)
    
def Wander(self, duration = 100):
  mouse_x = self.current_action.mouse_x
  mouse_y = self.current_action.mouse_y
  for i in range(int(duration)):
    mouse_x += min(1.0, max(-1.0, (self.random.random() - 0.5) / 20.0))
    mouse_y += min(1.0, max(-1.0, (self.random.random() - 0.5) / 20.0))
    yield Action(ACTION_WAIT, mouse_x, mouse_y)
    
def PushToKill(self, target, turbosity = -100.0):
  fx, fy, fz = self.transform_point(target.x, target.y, target.z, target.parent)
  dist = self.distance_to(target)
  best = None
  best_dist = 1000000.0
  target_front = Point(target, 0.0, 0.0, -15.0) % target.parent
  
  for pushable in self.level.pushables:
    if (not isinstance(pushable, Character)) and pushable.distance_to(target_front) < 55.0:
      px, py, pz = self.transform_point(pushable.x, pushable.y, pushable.z, pushable.parent)
      cur_dist = self.distance_to(pushable)
      if (cur_dist < dist) and (cur_dist < best_dist):
        best_dist = cur_dist
        best = pushable
        
  if not best: # Nothing to push
    if dist > 50.0: yield AttackRun(self, target)
    else:           yield Flee(self, target)
    return
  
  pushable = best
  while 1:
    if not pushable.parent: break
    v = target >> pushable
    v.set_length(pushable.radius * 2.0)
    t = pushable + v
    
    dist  = math.sqrt((t.x - self.x) ** 2 + (t.z - self.z) ** 2) #self.distance_to(t)
    distp = self.distance_to(pushable)
    if dist < 1.5: break
    
    if (not self.jumping) and (self.on_ground) and (distp < dist - 1.0) and (distp < pushable.radius * 2.5):
      self.jumping = 1
      yield Action(ACTION_JUMP, self.current_action.mouse_x, -1.0)
      
    yield goto(self, t, turbosity = turbosity)
    
  v = target >> pushable
  v.set_length(pushable.radius + self.radius - 0.1)
  t = pushable + v
  
  yield LookAt(self, t)
  
  for i in range(50):
    if pushable.worth_playing or not pushable.parent: break
    yield Action(ACTION_WAIT, 0.0, -1.0)
    
  if self.random.random() < 0.5:
    for i in range(10): yield Action(ACTION_WAIT, -1.0, 0.0)
  else:
    for i in range(10): yield Action(ACTION_WAIT,  1.0, 0.0)
  for i in range(10):   yield Action(ACTION_WAIT, 0.0, -1.0)
  

def LookAt(self, target):
  while 1:
    x, y, z = self.transform_point(target.x, target.y, target.z, target.parent)
    if abs(x / z) < 0.2: break
    if z < 0.0: yield Action(ACTION_WAIT, cmp(x, 0.0)) 
    else:       yield Action(ACTION_WAIT, cmp(x, 0.0))
    
Look_at = LookAt # old name
    
class Wait:
  def next(self): return Action()
WAIT = Wait()


class StackControler:
  def __init__(self, controlers  = None):
    self.controlers = controlers or []
    self.uncancelable_controler = None
    
  def animate(self, controler):
    self.controlers.append(controler)
    if not self.uncancelable_controler in self.controlers:
      self.uncancelable_controler = controler
      
  def append(self, controler):
    if not self.uncancelable_controler in self.controlers:
      self.controlers.append(controler)
      
  def next(self):
    try:
      c = self.controlers[-1]
      r = c.next() # calling c.next() may add a new controler ! => c may no longer be self.controlers[-1]
    except StopIteration:
      self.controlers.remove(c)
      if self.uncancelable_controler is c: self.uncancelable_controler = None
      return self.next()
    
    if not isinstance(r, Action):
      self.controlers.append(r)
      return self.next()
    
    return r

# A controler for phantom character, needed for them to play video sequences.
class PhantomControler(StackControler):
  def next(self):
    if self.controlers: return StackControler.next(self)


import soya.widget as widget, soya.opengl as soyaopengl

class KeyboardMouseControler(StackControler, widget.Widget):
  key_controlers = {}
  
  def __init__(self, character = None):
    StackControler.__init__(self)
    
    self.d_mouse_x     = self.d_mouse_y = 0.0
    self.mouse_x       = 0.0
    self.mouse_y       = 1.0
    self.character     = character
    self.enabled       = 1
    self.keyboard_mode = 0
    self.pixel_x = self.pixel_y = 0.0
    self.allow_move_back = 0
    
    from py2play.idler import IDLER
    IDLER.no_blackbands_group.add(self)
    IDLER.blackbands_group.add(self)
    
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.parent_left   = parent_left
    self.parent_top    = parent_top
    self.parent_width  = parent_width
    self.parent_height = parent_height

  if globdef.QUALITY > 1:
    def render(self):
      if not self.keyboard_mode:
        angle = math.atan(self.mouse_x / (self.mouse_y - 1.01))
        
        if self.allow_move_back < 10:
          size = 30.0 - self.mouse_y * 10.0
          soyaopengl.glEnable(soyaopengl.GL_BLEND)
          soyaopengl.glColor4f(0.0, 0.0, 0.0, 0.5)
          soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
          soyaopengl.glVertex2f(self.pixel_x - 9.0, self.pixel_y + 9.0)
          soyaopengl.glVertex2f(
            self.pixel_x - 9.0 - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + 9.0 + size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x - 9.0 + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + 9.0 + size * math.cos(angle) - size * math.sin(angle),
            )
          soyaopengl.glEnd()
          soyaopengl.glDisable(soyaopengl.GL_BLEND)

          soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
          soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
          soyaopengl.glVertex2f(self.pixel_x, self.pixel_y)
          soyaopengl.glVertex2f(
            self.pixel_x - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + size * math.cos(angle) - size * math.sin(angle),
            )
          soyaopengl.glEnd()
        else:
          size = 10.0 + self.mouse_y * 10.0
          soyaopengl.glEnable(soyaopengl.GL_BLEND)
          soyaopengl.glColor4f(0.0, 0.0, 0.0, 0.5)
          soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
          soyaopengl.glVertex2f(self.pixel_x - 9.0, self.pixel_y + 9.0)
          soyaopengl.glVertex2f(
            self.pixel_x - 9.0 + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + 9.0 - size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x - 9.0 - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + 9.0 - size * math.cos(angle) - size * math.sin(angle),
            )
          soyaopengl.glEnd()
          soyaopengl.glDisable(soyaopengl.GL_BLEND)

          soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
          soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
          soyaopengl.glVertex2f(self.pixel_x, self.pixel_y)
          soyaopengl.glVertex2f(
            self.pixel_x + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y - size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y - size * math.cos(angle) - size * math.sin(angle),
            )
          soyaopengl.glEnd()
  else:
    def render(self):
      if not self.keyboard_mode:
        angle = math.atan(self.mouse_x / (self.mouse_y - 1.01))
        
        soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
        soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
        soyaopengl.glVertex2f(self.pixel_x, self.pixel_y)
        
        if self.allow_move_back < 10:
          size = 30.0 - self.mouse_y * 10.0
          soyaopengl.glVertex2f(
            self.pixel_x - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y + size * math.cos(angle) - size * math.sin(angle),
            )
        else:
          size = 10.0 + self.mouse_y * 10.0
          soyaopengl.glVertex2f(
            self.pixel_x + size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y - size * math.cos(angle) + size * math.sin(angle),
            )
          soyaopengl.glVertex2f(
            self.pixel_x - size * math.cos(angle) + size * math.sin(angle),
            self.pixel_y - size * math.cos(angle) - size * math.sin(angle),
            )
        soyaopengl.glEnd()
        
  def next(self):
    character = self.character
    a = character.player.default_action
    
    if (character.player.default_action == ACTION_ROTUP) or (character.player.default_action == ACTION_ROTDOWN):
      if character.on_ground:
        character.player.default_action = a = ACTION_WAIT
        
    for event in soya.coalesce_motion_event(soya.process_event()):
      if   event[0] == soyasdlconst.MOUSEMOTION:
        self.keyboard_mode = 0
        camera = character.get_root().camera
        h = camera.get_screen_width ()
        w = camera.get_screen_height()
        self.pixel_x = float(event[1])
        self.pixel_y = float(event[2])
        self.mouse_x = ((float(event[1]) / camera.get_screen_width ()) - 0.5) * 2.0
        if self.allow_move_back < 10:
          self.mouse_y = ((float(event[2]) / camera.get_screen_height()) - 0.5) * 2.0
          if self.mouse_y > 0.95: self.allow_move_back += 2
        else:
          self.mouse_y = ((float(event[2]) / camera.get_screen_height()) * 3.0) - 1.0
          if self.mouse_y < 1.0: self.allow_move_back = 0
        
      elif self.enabled:
        if self.controlers and (event[0] != soyasdlconst.MOUSEBUTTONUP) and (not self.uncancelable_controler in self.controlers):
          self.controlers *= 0
          
        if event[0] == soyasdlconst.MOUSEBUTTONDOWN:
          if   event[1] == 1:
            if character.on_ground: a = ACTION_JUMP
            else:             character.player.default_action = a = ACTION_ROTUP
          elif event[1] == 2:
            if character.on_ground:
              character.player.controler.append(Wall_climber(character))
              return
            else: character.player.default_action = a = ACTION_ROTDOWN
          elif event[1] == 3: character.player.default_action = a = ACTION_TURBO
          elif event[1] == 4:
            if character.player.default_action == ACTION_ROTUP  : character.player.default_action = a = action.ACTION_WAIT
            else:                                                 character.player.default_action = a = ACTION_ROTDOWN
          elif event[1] == 5:
            if character.player.default_action == ACTION_ROTDOWN: character.player.default_action = a = action.ACTION_WAIT
            else:                                                 character.player.default_action = a = ACTION_ROTUP
            
        elif event[0] == soyasdlconst.MOUSEBUTTONUP:
          if   event[1] == 1: character.player.default_action = a = ACTION_WAIT
          elif event[1] == 3: character.player.default_action = a = ACTION_WAIT
          
        elif event[0] == soyasdlconst.KEYDOWN:
          if   (event[1] == soyasdlconst.K_LCTRL) or (event[1] == soyasdlconst.K_RCTRL):
            if character.on_ground: a = ACTION_JUMP
            else: character.player.default_action = a = ACTION_ROTUP
            
          elif (event[1] == soyasdlconst.K_LSHIFT) or (event[1] == soyasdlconst.K_RSHIFT):
            if character.on_ground:
              character.player.controler.append(Wall_climber(character))
              return
            else: character.player.default_action = a = ACTION_ROTDOWN
            
          elif (event[1] == soyasdlconst.K_LALT) or (event[1] == soyasdlconst.K_RALT):
            character.player.default_action = a = ACTION_TURBO
            
          elif event[1] == soyasdlconst.K_DOWN:  self.keyboard_mode = 1; self.d_mouse_y =  0.15
          elif event[1] == soyasdlconst.K_UP:    self.keyboard_mode = 1; self.d_mouse_y = -0.15
          elif event[1] == soyasdlconst.K_LEFT:  self.keyboard_mode = 1; self.d_mouse_x = -0.1
          elif event[1] == soyasdlconst.K_RIGHT: self.keyboard_mode = 1; self.d_mouse_x =  0.1
          
          elif (event[1] == soyasdlconst.K_q) or (event[1] == soyasdlconst.K_ESCAPE):
            from py2play.idler import IDLER
            IDLER.end_game()
            
          elif event[1] == soyasdlconst.K_s:
            import tempfile
            screenshot_file = tempfile.mktemp(".jpeg")
            soya.screenshot(screenshot_file)
            print "* Slune * Screenshot saved as %s" % screenshot_file
            
          elif event[1] == soyasdlconst.K_w:
            soya.toggle_wireframe()

          elif event[1]:
            Controler = self.key_controlers.get(event[1])
            if Controler:
              character.player.controler.append(Controler(character))
              
        elif event[0] == soyasdlconst.KEYUP:
          if   event[1] == soyasdlconst.K_DOWN:
            if self.d_mouse_y > 0.0: self.d_mouse_y = 0.0
            if self.mouse_y == 1.0: self.allow_move_back = 10
          elif event[1] == soyasdlconst.K_UP:
            if self.d_mouse_y < 0.0: self.d_mouse_y = 0.0
          elif event[1] == soyasdlconst.K_LEFT:
            if self.d_mouse_x < 0.0: self.d_mouse_x = 0.0
          elif event[1] == soyasdlconst.K_RIGHT:
            if self.d_mouse_x > 0.0: self.d_mouse_x = 0.0
            
          elif (event[1] == soyasdlconst.K_LALT) or (event[1] == soyasdlconst.K_RALT) or (event[1] == soyasdlconst.K_LCTRL) or (event[1] == soyasdlconst.K_RCTRL) or (event[1] == soyasdlconst.K_LSHIFT) or (event[1] == soyasdlconst.K_RSHIFT):
            character.player.default_action = a = ACTION_WAIT
            
        elif event[0] == soyasdlconst.JOYAXISMOTION:
          if event[1] == 0: # X Axis
            self.d_mouse_x = float(event[2]) / 327680
            self.keyboard_mode = 1
          elif event[1] == 1:   # Y Axis
            self.d_mouse_y = float(event[2]) / 327680
            self.keyboard_mode = 1
            
        elif event[0] == soyasdlconst.JOYBUTTONDOWN:
          if event[1] == 0:
            character.player.default_action = a = ACTION_TURBO
          elif event[1] == 1:
            if character.on_ground: a = ACTION_JUMP
            else: character.player.default_action = a = ACTION_ROTUP
          elif event[1] == 2:
            if character.on_ground:
              character.player.controler.append(Wall_climber(character))
              return
            else: character.player.default_action = a = ACTION_ROTDOWN

        elif event[0] == soyasdlconst.JOYBUTTONUP:
          character.player.default_action = a = ACTION_WAIT

      elif (event[0] == soyasdlconst.JOYBUTTONDOWN) or (event[0] == soyasdlconst.MOUSEBUTTONDOWN) or ((event[0] == soyasdlconst.KEYDOWN) and (not event[1] in (0, soyasdlconst.K_UP, soyasdlconst.K_DOWN, soyasdlconst.K_LEFT, soyasdlconst.K_RIGHT))):
        return Action(ACTION_STARTGAME)
      
    from py2play.idler import IDLER
    if self.controlers:
      if self.uncancelable_controler and not IDLER.blackbands_visible: IDLER.show_blackbands()
      
      return StackControler.next(self)
    
    else:
      if IDLER.blackbands_visible: IDLER.hide_blackbands()
      
      if self.keyboard_mode:
        self.mouse_x *= 0.9
        if self.mouse_y < 0.0: self.mouse_y *= 0.9
        
        self.mouse_x += self.d_mouse_x
        if   self.mouse_x < -1.0: self.mouse_x = -1.0
        elif self.mouse_x >  1.0: self.mouse_x =  1.0
        
        self.mouse_y += self.d_mouse_y
        if   self.mouse_y < -1.0: self.mouse_y = -1.0
        else:
          if self.allow_move_back < 10:
            if self.mouse_y > 1.0: self.mouse_y =  1.0
          else:
            if   self.mouse_y < 1.0: self.allow_move_back = 0
            elif self.mouse_y > 2.0: self.mouse_y =  2.0
            
      return Action(a, self.mouse_x, self.mouse_y)
    
    
    
#KeyboardMouseControler.key_controlers[soyasdlconst.K_z] = Push_nearest_character #z
#KeyboardMouseControler.key_controlers[soyasdlconst.K_e] = Wall_climber #e
#KeyboardMouseControler.key_controlers[soyasdlconst.K_f] = Goto_next_flag #f
KeyboardMouseControler.key_controlers[soyasdlconst.K_g] = Dodge #g



          



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import struct, math, random
import py2play.action as action, py2play.character as py2play_character
#from py2play import recvall
import soya, soya.ray as ray, soya.widget as widget

from soya import Point, Vector
import slune.globdef as globdef, slune.sound as sound

ACTION_WAIT      = action.ACTION_WAIT
ACTION_JUMP      = "j"
ACTION_TURBO     = "t"
ACTION_ROTUP     = "u"
ACTION_ROTDOWN   = "d"
ACTION_ROTLAT    = "l"
ACTION_DEBUG     = "°"
ACTION_STARTGAME = "!"

class Action:
  def __init__(self, action = ACTION_WAIT, mouse_x = 0.0, mouse_y = 1.0, round = None):
    self.round   = round
    self.action  = action
    self.mouse_x = mouse_x
    self.mouse_y = mouse_y
    
  def __repr__(self):
    return "<Action %s at round %s, (%s, %s)>" % (self.action, self.round, self.mouse_x, self.mouse_y)
  

from socket import MSG_WAITALL


#def WRITE_ACTION(socket, action):
#  socket.sendall(struct.pack("!cddi", action.action, action.mouse_x, action.mouse_y, action.round))
  
#def READ_ACTION(socket):
#  return Action(*struct.unpack("!cddi", recvall(socket, 21)))

action.Action       = Action
#action.READ_ACTION  = READ_ACTION
#action.WRITE_ACTION = WRITE_ACTION

RIGHT  = Vector(None, 1.0,  0.0,  0.0)
DOWN   = Vector(None, 0.0, -1.0,  0.0)
UP     = Vector(None, 0.0,  1.0,  0.0)
FRONT  = Vector(None, 0.0,  0.0, -1.0)

# Re-useable Points and Vectors

_P  = Point()
_P2 = Point()
_P3 = Point()
_V  = Vector()
_V2 = Vector()

VEHICLE_PERSO_COORDS = [
  (0.0, 0.1 , 0.03),
  (0.0, 0.1 , 0.03),
  (0.0, 0.4 , -0.2),
  (0.0, 0.3 , 0.35),
  (0.0, 0.48, 0.0),
  ]

class Pusher: pass
class Competitor: pass

class Character(soya.World, Pusher):
  nitro_enabled  = 1
  push_fly       = 0.3
  context        = None
  
  def __init__(self, parent = None):
    soya.World.__init__(self, parent)
    
    self.radius      = 0.7
    self.floating    = 0.3
    self.weight      = 1.0
    self.rotup_speed = 7.0
    self.mode_grip   = 0
    self.gripped     = 0
    self.rot_vert = 0.0
    self.speed  = Vector(self)
    self.force  = Vector(None)
    self.solid  = 0
    self.nitro = None
    self.nitro_time = 0
    self.flame_thrower = 0
    self.flame_throw_time = 0
    
    self.camera_target = self
    self.old_speed_x = 0.0
    self.context = None
    self.on_ground = 0
    self.was_on_ground = 0
    self.vertical_angle = 0.0
    self.platform_speed = None
    
    self.CENTER = Point(self, 0.0, 0.3,  0.0)
    
    self.scramble = 0.0
    self.next_pos = Point(self, 0.0, 0.0, 0.0)
    
    self.internal = soya.World(self)
    self.perso    = soya.Volume(self.internal)
    
    self.perso_name = None
    self.vehicle = 1
    self.life  = 1.0
    self.pushed_by = None
    
    # IA stuff
    self.jumping   = 0.0
    self.max_speed = 1.0
    self.current_action = Action()
    self.speed.y = 0.01 # Avoid null speed (null speed => jumping for AI). The speed will be overrided later.
    self.check_ground_vector = Vector(self, 0.0, -1.0, -2.0)
    
    # blam hack to test raypicking
    #import soya.laser
    #laser = soya.laser.Laser(self)
    #laser.rotate_vertical(-90.0)
    #laser.reflect = 1
    
    
  def __getstate__(self):
    state = soya.World.__getstate__(self)
    if state[1] is self.__dict__: state = state[0], state[1].copy()
    del state[1]["context"]
    return state
  def __setstate__(self, state):
    state = soya.World.__setstate__(self, state)
    self.context = None
    
  def set_perso(self, perso_name):
    self.perso_name = perso_name
    self.perso.set_shape(soya.Shape.get(self.perso_name + "1"))
    
  def displayname(self): return self.perso_name.title()
  
  def set_vehicle(self, vehicle, add_ray = 1):
    self.vehicle = vehicle
    self.internal.set_shape(soya.Shape.get("vehicle" + str(self.vehicle)))
    if hasattr(self, "ray"):
      self.internal.remove(self.ray)
      del self.ray
    if hasattr(self, "ray2"):
      self.internal.remove(self.ray2)
      del self.ray2
      
    if   vehicle == 0:
      self.radius      = 0.7
      self.floating    = 0.5
      self.weight      = 1.3
      self.maxspeedx   = 0.8
      self.maxspeedz   = 0.98
      self.rotup_speed = 7.0
      self.mode_grip   = 0
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray.set_xyz(-0.84, 0.10, 0.0)
        self.ray.endpoint.set_xyz(0.45, 0.0, 0.0)
        self.ray.length = 20
        
        self.ray2 = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray2.set_xyz(0.84, 0.10, 0.0)
        self.ray2.endpoint.set_xyz(-0.45, 0.0, 0.0)
        self.ray2.length = 20
        
    elif vehicle == 1:
      self.radius      = 0.7
      self.floating    = 0.3
      self.weight      = 1.0
      self.maxspeedx   = 1.0
      self.maxspeedz   = 1.0
      self.rotup_speed = 8.0
      self.mode_grip   = 0
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL)
        self.ray.set_xyz(-0.3, 0.15, 0.5)
        self.ray.endpoint.set_xyz(0.6, 0.0, 0.0)
        self.ray.length = 20
        
    elif vehicle == 2:
      self.radius      = 0.8
      self.floating    = 0.3
      self.weight      = 1.6
      self.maxspeedx   = 0.9
      self.maxspeedz   = 0.95
      self.rotup_speed = 6.0
      self.mode_grip   = 0
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray.set_xyz(-0.5, 0.2, 1.0)
        self.ray.endpoint.set_xyz(1.0, 0.0, 0.0)
        self.ray.length = 20
        
    elif vehicle == 3:
      self.radius      = 0.7
      self.floating    = 0.3
      self.weight      = 0.6
      self.maxspeedx   = 1.0
      self.maxspeedz   = 1.0
      self.rotup_speed = 9.0
      self.mode_grip   = 1
      self.DOWN = Vector(self, 0.0, -1.0, 0.0)
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL2)
        self.ray.set_xyz(0.0, 0.07, 0.65)
        self.ray.endpoint.set_xyz(0.0, 0.3, 0.0)
        self.ray.length = 20
        
    elif vehicle == 4:
      self.radius      = 1.0
      self.floating    = 0.5
      self.weight      = 2.2
      self.maxspeedx   = 0.7
      self.maxspeedz   = 0.8
      self.rotup_speed = 4.0
      self.mode_grip   = 0
      self.DOWN = Vector(self, 0.0, -1.0, 0.0)
      
      if add_ray:
        self.ray = ray.Ray(self.internal, RAY_MATERIAL3)
        self.ray.set_xyz(-0.6, 0.35, 0.6)
        self.ray.endpoint.set_xyz(1.2, 0.0, 0.0)
        self.ray.set_xyz(-0.95, -0.4, 0.0)
        self.ray.endpoint.set_xyz(1.9, 0.0, 0.0)
        self.ray.length = 10
        
    self.perso.set_xyz(*VEHICLE_PERSO_COORDS[vehicle])
    self.perso_y = self.perso.y
    
  def add_headlight(self):
    light = self.light = soya.Light(self)
    light.set_xyz(0.0, 1.2, -0.5)
    light.diffuse     = (0.5, 0.6, 1.0, 1.0)
    light.specular    = (0.0, 0.0, 0.0, 1.0)
    light.quadratic   = 0.0
    light.linear      = 0.07
    light.constant    = 0.0
    light.exponent    = 3.0
    light.angle       = 90.0
    light.cast_shadow = 0
    
  def init(self): pass
  
  def default_mouse_x(self): return 0.0
    
  def set_level(self, level, init = 0):
    level.add(self)
    
    super(Character, self).set_level(level, init)
    
  def begin_action(self, action):
    from py2play.idler import IDLER as root
    
    speed = self.speed
    force = self.force
    
    self.internal.rotate_incline((self.old_speed_x - speed.x) *  6000.0)
    self.perso   .rotate_lateral((self.old_speed_x - speed.x) * 12000.0)
    self.old_speed_x = speed.x
    self.turn_lateral(-500.0 * speed.x)
    
    if self.flame_throw_time: self.flame_throw_time -= 1
    if self.nitro_time:
      self.nitro.set_sizes((
        1.0 - (10 - self.nitro_time) / 15.0,
        1.0 - (10 - self.nitro_time) / 15.0,
        ))
      self.nitro_time -= 1
      if not self.nitro_time:
        self.nitro.auto_generate_particle = 0
        
    self.life = min(1.0, self.life + 0.0003)
    
    if   action.action == ACTION_ROTUP  : self.turn_vertical( self.rotup_speed)
    elif action.action == ACTION_ROTDOWN: self.turn_vertical(-self.rotup_speed)
    elif action.action == ACTION_ROTLAT : self.rotate_lateral(30.0)
    elif (action.action == ACTION_TURBO) and (self.flame_throw_time == 0):
      if self.flame_thrower and ((not self.nitro) or not self.nitro.auto_generate_particle):
        self.flame_thrower -= 1
        self.flame_throw_time = 50
        flame_thrower = FlameThrower(self)
        
      elif self.nitro_enabled:
        length = force.length()
        d = 0.09 * max(0.0, (1.0 - length))
        if self.life > d:
          force.add_vector(Vector(self, 0.0, 0.0, -2.0) % root)
          force.set_length(max(1.0, length))
          self.life -= 0.08 * max(0.0, (1.0 - length))
          if self.nitro is None:
            self.nitro = Nitro(self)
            sound.play("nitro-1.wav", self)
          elif not self.nitro.auto_generate_particle:
            self.nitro.auto_generate_particle = 1
            sound.play("nitro-1.wav", self)
          self.nitro.set_sizes((1.0, 1.0))
          self.nitro_time = 10
          
    elif action.action == ACTION_DEBUG  :
      print self.position(), self.round
      print self.level.parent.camera.position(), Point(self.level.parent.camera, 0.0, 0.0, -1.0) % self.get_root()
      self.hurt(2.0)
      poqjfcpjk
      
    elif (action.action == ACTION_JUMP) and self.was_on_ground:
      force.y       += 0.4
      self.rot_vert += 0.4
      
    if self.context:
      self.was_on_ground = self.on_ground
      
      if   self.context.raypick(self.CENTER, DOWN, self.floating + 0.45, 1, 1, _P, _V):
        self.on_ground = 1
        self.gripped   = 0
      elif self.mode_grip and self.context.raypick(self.CENTER, self.DOWN, 6.0 * self.speed.length(), 1, 1, _P, _V):
        self.on_ground = self.gripped = 1
      else:
        self.on_ground = self.gripped = 0

    if self.context and self.on_ground:
      _P.convert_to(self.parent)
      
      if (self.y < _P.y + self.floating - force.y):
        if (not self.scramble) and self.parent:
          self.scramble = min(30.0, ((force.y + self.parent.transform_vector(speed.x, speed.y, speed.z, self)[1]) * 7.0) ** 2)
          
          #if (self.scramble > 2) and (-20.0 < self.y < 20.0):
          #  Snow(self.level).move(self)

        # "force.y = _P.y + self.floating - self.y" give better visual effect, but clashes
        # with mode grip.
        if self.gripped: force.y = 0.0
        else: force.y = _P.y + self.floating - self.y
        
      elif self.y > _P.y + self.floating + 0.1: force.y -= 0.03
      
      if force.y < 0.2:
        if (action.action != ACTION_ROTUP) and (action.action != ACTION_ROTDOWN):
          _V.convert_to(self)
          
          if _V.y < -0.3: self.turn_incline(10.0)
          elif (_V.y < 0.996) and _V.y:
            #soya.cross_vector(_V, UP, _V2)
            _V.cross_product(UP, _V2)
            _V2.convert_to(self.parent)
            self.vertical_angle = -_V.angle_to(UP)
            self.rotate_axe(self.vertical_angle / 3.0, _V2)
            #self.rotate_axe(-_V.angle_to(UP) / 4.0, _V2.x, _V2.y, _V2.z)
    else:
      force.y -= 0.03 # comment this line if you want to fly
      if self.y < -35.0: self.hurt(1.0)
      
    force.__imul__(0.95)
    
    if self.rot_vert > 0.0:
      self.turn_vertical(self.rot_vert * 14.0)
      self.rot_vert -= 0.03
      
    x, y, z = self.transform_vector(force.x, force.y, force.z, self.parent)
    speed.x += x
    speed.y += y
    speed.z += z
    
    if speed.length() > 0.75: speed.set_length(0.75)
    
    for character in self.level.pushables:
      if not character is self:
        if self.distance_to(character) < self.radius + character.radius:
          _V.set_start_end(self, character)
          
          f = 3.0 * max(0.0, _V.dot_product(self.speed)) * self.weight / character.weight
          if f:
            _V.set_length(f)
            character.push(_V.x, _V.y + character.push_fly * f, _V.z, self)
            f = -0.3 * character.weight / self.weight
            self.push(f * _V.x, f * _V.y, f * _V.z, character)
            
    _P2.__init__(self, 0.0, 0.3, 0.0)
    _P2.add_vector(speed)
    _P2.convert_to(root)

    self.context = root.RaypickContext(_P2, max(self.radius, 0.7) + 0.5, self.context)
    bouree = 0
    
    
    for vec in (RIGHT, FRONT):
      if self.context.raypick(_P2, vec, self.radius, 0, 0, _P, _V):
        _P.convert_to(root)
        _V.convert_to(root)
        
        #a = _V.cross_product(Vector(self, 0.0, 1.0, 0.0)) % root
        #print a
        #self.turn_axe(-15.0, a.x, a.y, a.z)
        
        if vec.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * vec.x, _P2.y - self.radius * vec.y, _P2.z - self.radius * vec.z)
        else:                         _P3.__init__(root, _P2.x + self.radius * vec.x, _P2.y + self.radius * vec.y, _P2.z + self.radius * vec.z)
        
        if force.dot_product(_V) < 0.0:
          if abs(force.x) + abs(force.z) > 0.1: # Else force is only due to gravity!
            _V2.clone(_V)
            _V2.set_length(-2.0 * force.dot_product(_V2))
            force += _V2
            force *= 0.5
            
          bouree = 1
          
        if (_P.x - _P2.x) * _V.x + (_P.y - _P2.y) * _V.y + (_P.z - _P2.z) * _V.z > 0.0:
          _V.set_start_end(_P2, _P)
        else: _V.set_start_end(_P3, _P)
        
        _P2  .__iadd__(_V)
        speed.__iadd__(_V)
        
    _P2.y += 0.2
    if self.context.raypick(_P2, DOWN, 0.7, 0, 1, _P, _V):
      _P.convert_to(root)
      _V.convert_to(root)
      if DOWN.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * DOWN.x, _P2.y - self.radius * DOWN.y, _P2.z - self.radius * DOWN.z)
      else:
        if force.y > 0.0: force.y = 0.0 # Stop and break a jump
        _P3.__init__(root, _P2.x + self.radius * DOWN.x, _P2.y + self.radius * DOWN.y, _P2.z + self.radius * DOWN.z)
      if force.dot_product(_V) < -0.4:
        bouree = 1
      
      _V.set_start_end(_P3, _P)
      
      #_P2   .__iadd__(_V) # Useless since it is the last raypicking.
      speed.__iadd__(_V)
      
    #if bouree and not self.nitro.auto_generate_particle:
    if bouree:
      couic = min(0.5, force.x ** 2 + force.z ** 2) * 0.8
      if couic > 0.01:
        #force.x = force.z = 0.0
        
        if self.hurt(couic) and (len(explosions) < 4): # Limit the number of simultaneous explosions
          explode = Explosion(self.parent)
          explode.move(self)
          sound.play("explose-1.wav", explode)
          
    self.next_pos.clone(speed)
    self.next_pos.convert_to(self.parent)
    self.next_pos.parent = None
    
  def push(self, x, y, z, pusher = None, play_sound = 1):
    if play_sound: sound.play("push-1.wav", self)
    
    from py2play.idler import IDLER
    def f():
      self.force.add_xyz(x, y, z)
      self.rot_vert += max(0.0, y)
    IDLER.next_round_tasks.append(f)
    
    self.pushed_by = pusher
    
  def end_round(self):
    # Restore the real position -- advance_time may cause rounding error on the position !!!
    self.move(self.next_pos)
    if self.platform_speed:
      self += self.platform_speed
      self.platform_speed = None
      
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    
    self.add_mul_vector(proportion, self.speed)
    
    if self.scramble:
      if self.scramble <= proportion:
        self.internal.y = 0.0
        self.scramble   = 0.0
      else:
        self.internal.y = abs(math.sin(self.scramble)) * self.scramble * 0.04
        self.scramble  -= proportion / 3.0
      self.perso.y = self.perso_y + self.internal.y
      
  def teleport(self, x, y, z):
    self.set_xyz(x, y, z)
    self.force.set_xyz(0.0, 0.0, 0.0)
    if self.scramble:
      self.scramble   = 0.0
      self.internal.y = 0.0
      self.perso.y    = self.perso_y
    if hasattr(self, "ray" ): self.ray .zap()
    if hasattr(self, "ray2"): self.ray2.zap()
    self.context = None # The raypick context is no longer valid !!!
    
  def hurt(self, life):
    """Kills LIFE life to this character. Returns true if the character is still alive."""
    self.life -= life
    if self.life < 0.0: self.die()
    else: return 1
    
  def die(self): 
    explode = Explosion(self.parent)
    explode.move(self)
    explode.life = 60
    explode.particle_width = explode.particle_height = 0.7
    
    sound.play("explose-3.wav", explode)
    self.level.remove_character(self)

class NonPlayerCharacter(Character, py2play_character.NonPlayerCharacter):
  clan = 0
  
  def __init__(self, parent = None):
    py2play_character.NonPlayerCharacter.__init__(self)
    Character.__init__(self, parent)
    
    import slune.controler
    self.controler = slune.controler.StackControler([self.IAControler()])
    
  def choose_action(self):
    return self.controler.next()
  
  def begin_round(self):
    py2play_character.NonPlayerCharacter.begin_round(self)
    Character.begin_round(self)
    
  def begin_action(self, action):
    self.speed.set_xyz(action.mouse_x / 120.0 * self.maxspeedx, 0.0, (action.mouse_y * 0.2 - 0.2) * self.maxspeedz)
    
    Character.begin_action(self, self.current_action)
    
  def __getstate__(self):
    state = Character.__getstate__(self)
    del state[0]["controler"]
    return state
  def __setstate__(self, state):
    state = Character.__setstate__(self, state)
    import slune.controler
    self.controler = slune.controler.StackControler([self.IAControler()])
    
  def IAControler(self):
    while 1: yield Action() # Does nothing !
    
  
class Pushable(soya.Volume, Pusher):
  clan           = 0
  bonus_chance   = 0.15
  bonus_type     = 0
  on_bourre      = None
  push_fly       = 0.3
  def __init__(self, parent = None):
    soya.Volume.__init__(self, parent)
    
    self.radius      = 1.0
    self.radius_y    = 1.0
    self.weight      = 1.0
    self.life        = 1.0
    self.on_ground   = 1 # 0 ???
    
    self.speed       = Vector(self)
    self.force       = Vector(None)
    self.solid       = 0
    self.worth_playing = 0
    
    self.CENTER = Point(self, 0.0, 0.3,  0.0)
    
    self.next_pos = Point(self, 0.0, 0.0, 0.0)
    
  def begin_round(self):
    soya.Volume.begin_round(self)
    
    force = self.force
    if not self.worth_playing:
      if (abs(force.x) > 0.1) or (abs(force.y) > 0.1) or (abs(force.z) > 0.1):
        self.worth_playing = 1
        if self.on_bourre:
          if self.on_bourre == "win":
            from videosequence import MissionCompleted, PLAYER
            MissionCompleted(1).start(tux = PLAYER)
            
            from py2play.idler import IDLER
            IDLER.camera.add_traveling(soya.ThirdPersonTraveling(self))
            
          else: self.on_bourre()
      else: return
      
    speed = self.speed
    speed.set_xyz(0.0, 0.0, 0.0)
    
    from py2play.idler import IDLER as root
    
    if root.raypick(self.CENTER, DOWN, self.radius_y / 2.0 + 0.45, 1, 1, _P, _V):
      _P.convert_to(self.parent)
      self.on_ground = 1
      
      if (self.y < _P.y + self.radius_y / 2.0 - self.force.y):
        force.y = 0.0
        self.force.y += _P.y + self.radius_y / 2.0 - self.y
      elif self.y > _P.y + self.radius_y / 2.0 + 0.1: force.y -= 0.03
      
      if self.force.y < 0.05:
        _V.convert_to(self)
        
        if _V.y < 0.0: self.turn_incline(10.0)
        elif _V.y < 0.999:
          _V.cross_product(UP, _V2)
          _V2.convert_to(self.parent)
          angle = _V.angle_to(UP)
          self.rotate_axe(-angle / 4.0, _V2)
        else:
          self.worth_playing = 0
          
    else:
      self.on_ground = 0
      force.y -= 0.03
      if self.y < -35.0: self.die()
      
    force.__imul__(0.95)
    
    if force.y > 0.0: self.turn_vertical(force.y * 14.0)
    
    x, y, z = self.transform_vector(force.x, force.y, force.z, self.parent)
    speed.x += x
    speed.y += y
    speed.z += z
    
    if speed.length() > 0.5: speed.set_length(0.5)
    
    if self.parent:
      for character in self.parent.pushables:
        if not character is self:
          if self.distance_to(character) < self.radius + character.radius:
            _V.set_start_end(self, character)

            _V.y += character.push_fly
            f = 2.5 * max(0.0, _V.dot_product(self.speed))
            if f:
              _V.set_length(f * self.weight / character.weight)
              character.push(_V.x, _V.y, _V.z, self)
            
    _P2.__init__(self, 0.0, self.radius_y / 2.0, 0.0)
    _P2.add_vector(speed)
    _P2.convert_to(root)
    
    context = root.RaypickContext(_P2, max(self.radius, 0.7))
    
    bouree = 0
    
    for vec in (RIGHT, FRONT):
      if context.raypick(_P2, vec, self.radius, 0, 0, _P, _V):
        _P.convert_to(root)
        _V.convert_to(root)
        if vec.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius * vec.x, _P2.y - self.radius * vec.y, _P2.z - self.radius * vec.z)
        else:                        _P3.__init__(root, _P2.x + self.radius * vec.x, _P2.y + self.radius * vec.y, _P2.z + self.radius * vec.z)
        
        if force.dot_product(_V) < 0.0:
          _V2.clone(_V)
          _V2.set_length(-2.0 * force.dot_product(_V2))
          force += _V2
          #force *= 0.5
          
          bouree = 1
          
        if (_P.x - _P2.x) * _V.x + (_P.y - _P2.y) * _V.y + (_P.z - _P2.z) * _V.z > 0.0:
          _V.set_start_end(_P2, _P)
        else: _V.set_start_end(_P3, _P)
        
        _P2  .__iadd__(_V)
        speed.__iadd__(_V)
        
    _P2.y += 0.2
    if context.raypick(_P2, DOWN, self.radius_y, 0, 1, _P, _V):
      _P.convert_to(root)
      _V.convert_to(root)
      if DOWN.dot_product(_V) > 0.0: _P3.__init__(root, _P2.x - self.radius_y * DOWN.x, _P2.y - self.radius_y * DOWN.y, _P2.z - self.radius_y * DOWN.z)
      else:
        if force.y > 0.0: force.y = 0.0 # Stop and break a jump
        _P3.__init__(root, _P2.x + self.radius_y * DOWN.x, _P2.y + self.radius_y * DOWN.y, _P2.z + self.radius_y * DOWN.z)
        
      if force.dot_product(_V) < 0.0: bouree = 1
      
      _V.set_start_end(_P3, _P)
      
      #_P2   .__iadd__(_V) # Useless since it is the last raypicking.
      speed.__iadd__(_V)
      
    if bouree:
      couic = min(0.5, force.x ** 2 + force.z ** 2)
      if couic > 0.01:
        #force.x = force.z = 0.0
        
        if self.hurt(couic) and (len(explosions) < 4): # Limit the number of simultaneous explosions
          explode = Explosion(self.parent, nb_particles = 15)
          sound.play("explose-1.wav", explode)
          explode.move(self)
          
    self.next_pos.clone(speed)
    self.next_pos.convert_to(self.parent)
    self.next_pos.parent = None
    
  def push(self, x, y, z, pusher = None, play_sound = 1):
    if play_sound: sound.play("push-2.wav", self)
    from py2play.idler import IDLER
    IDLER.next_round_tasks.append(lambda : self.force.add_xyz(x, y, z))
    
  def end_round(self):
    if self.worth_playing:
      # Restore the real position -- advance_time may cause rounding error on the position !!!
      self.move(self.next_pos)
      
  def advance_time(self, proportion):
    if self.worth_playing:
      self.add_mul_vector(proportion, self.speed)
      
  def hurt(self, life):
    """Kills LIFE life to this character. Returns true if the character is still alive."""
    
    self.life -= life
    if self.life < 0.0: self.die()
    else: return 1
    
  def die(self):
    explode = Explosion(self.parent, nb_particles = 20)
    explode.move(self)
    explode.life = 30
    explode.particle_width = explode.particle_height = 0.7
    
    if self.bonus_type:
      import slune.level as slune_level
      if   self.bonus_type == 1: bonus = slune_level.LifeBonus(self.parent)
      elif self.bonus_type == 2: bonus = slune_level.FlameThrowerBonus(self.parent)
      bonus.move(self)
      bonus.come_back = 0
      self.parent.bonuss.append(bonus)
      sound.play("explose-2.wav", explode)
    else:
      sound.play("explose-2.wav", explode)
    
    self.parent.pushables.remove(self)
    self.parent.remove(self)
    
def Barrier(parent, p = None):
  p = p or Pushable(parent)
  p.set_shape(soya.Shape.get("gate3"))
  p.radius_y = 0.5
  p.life     = 0.5
  p.weight   = 0.6
  return p
def Barrier2(parent, p = None):
  p = p or Pushable(parent)
  p.set_shape(soya.Shape.get("gate4"))
  p.radius_y = 0.5
  p.life     = 0.5
  p.weight   = 0.6
  return p
def TreeTrunc(parent, p = None):
  p = p or Pushable(parent)
  p.set_shape(soya.Shape.get("tree_trunc"))
  p.radius   = 2.0
  p.radius_y = 0.3
  p.life     = 2.0
  p.weight   = 1.2
  return p
def Quille(parent, p = None):
  p = p or Pushable(parent)
  p.set_shape(soya.Shape.get("quille-1"))
  p.radius   = 0.7
  p.radius_y = 0.7
  p.life     = 0.1
  p.weight   = 0.4
  return p
def Ball(parent, p = None):
  p = p or Pushable(parent)
  p.set_shape(soya.Shape.get("ball1"))
  p.radius   = 0.9
  p.radius_y = 0.9
  p.life     = 50.0
  p.weight   = 0.1
  p.bonus_chance = 0.0
  return p


class TakeableVehicle(soya.Volume):
  radius = 5.0
  
  def __init__(self, parent = None, vehicle = 3):
    soya.Volume.__init__(self, parent)
    self.set_vehicle(vehicle)
    self.disabled = {}
    self.changing = 0
    self.solid    = 0
    
  def set_vehicle(self, vehicle):
    self.vehicle = vehicle
    self.set_shape(soya.Shape.get("vehicle" + str(self.vehicle)))
    
  def disable(self, character):
    self.disabled[character] = 1
    self.changing = 0
    
  def begin_round(self):
    soya.Volume.begin_round(self)
    
    if not self.changing:
      for character in self.parent.characters:
        if isinstance(character, PlayerCharacter):
          if self.disabled.get(character, 0):
            if (self.distance_to(character) > self.radius * 2.0): del self.disabled[character]
          elif (self.distance_to(character) < self.radius):
            self.changing = 1 # Avoid 2 changing at the same time !!!
            from videosequence import ChangeVehicle
            ChangeVehicle(character, self).start(character = character)
            
class Remorque(NonPlayerCharacter, py2play_character.NonPlayerCharacter):
  clan = 0
  
  def __init__(self, parent = None, remorqued_by = None, remorqued_distance = 1.5):
    py2play_character.NonPlayerCharacter.__init__(self)
    NonPlayerCharacter.__init__(self, parent, 0)
    
    self.remorqued_by       = remorqued_by
    self.remorqued_distance = remorqued_distance
    
    self.set_vehicle(1)
    
  def begin_action(self, action):
    self.current_action = Action()
    self.speed.set_xyz(0.0, 0.0, 0.0)
    
    if self.remorqued_by:
      _V.set_start_end(self, self.remorqued_by)
      _V.convert_to(self)
      _V.set_length(_V.length() - self.remorqued_distance)
      if _V.length() > 0.48: _V.set_length(0.48)
      _V.y = _V.y - self.force.y
      self.speed.set_xyz(_V.x / 20.0, _V.y, _V.z)
    else:
      self.speed.set_xyz(0.0, 0.0, 0.0)
      
    Character.begin_action(self, self.current_action)
    
RAY_MATERIAL = soya.Material()
RAY_MATERIAL.diffuse = (0.7, 0.4, 0.95, 1.0)
RAY_MATERIAL.additive_blending = 1

RAY_MATERIAL2 = soya.Material()
RAY_MATERIAL2.diffuse = (0.9, 0.6, 0.3, 1.0)
RAY_MATERIAL2.additive_blending = 1

RAY_MATERIAL3 = soya.Material()
RAY_MATERIAL3.diffuse = (0.0, 0.0, 0.9, 1.0)
RAY_MATERIAL3.additive_blending = 1

def PlayerLight(parent):
  return None
  light = soya.Light(parent)
  light.set_xyz(0.0, 1.2, -0.5)
  light.diffuse   = (0.5, 0.6, 1.0, 1.0)
  light.specular  = (0.0, 0.0, 0.0, 1.0)
  light.quadratic = 0.005
  light.linear    = 0.2
  light.constant  = 0.0
  light.exponent  = 1.0
  
  light.look_at(Vector(parent, 0.0, 0.0, -1.0))
  
  light.angle = 90.0
  
  return light

class PlayerCharacter(Character, py2play_character.PlayerCharacter, Competitor):
  clan = 1
  
  def __init__(self, player = None):
    py2play_character.PlayerCharacter.__init__(self, player)
    Character.__init__(self)
    
    if player: player.default_action = action.ACTION_WAIT
    
    self.set_perso  (globdef.CHARACTER.lower())
    self.set_vehicle(globdef.VEHICLE)
    
    #self.light = PlayerLight(self)
    
    if globdef.VERIFICATION_SERVER and self.player: self.logon_verif()
    
  def displayname(self): return self.perso_name.title() + " (" + self.player.name + ")"
  
  def default_mouse_x(self):
    if self.controler.uncancelable_controler: return 0.0
    return self.player.controler.mouse_x
  
  def logon_verif(self):
    """Log on the verif server."""
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((globdef.VERIFICATION_SERVER, 36081))
    self.player.verif = s.makefile("w", 0)
    self.player.verif.write("P") # Check position
    
  def begin_round(self):
    if self.player.active: self.plan_action(self.player.controler.next())
    else:                  self.controler.next()
    
    py2play_character.PlayerCharacter.begin_round(self)
    Character.begin_round(self)
    
  def begin_action(self, action):
    if globdef.VERIFICATION_SERVER:
      import cPickle as pickle
      pickle.dump((self.player.name, self.round, (self.x, self.y, self.z), None), self.player.verif)
      
    # Compute the speed
    #self.speed.set_xyz(action.mouse_x / 120.0 * self.maxspeedx, 0.0, (action.mouse_y * 0.2 - 0.28)  * self.maxspeedz)
    self.speed.set_xyz(action.mouse_x / 120.0 * self.maxspeedx, 0.0, (action.mouse_y * 0.2 - 0.2)  * self.maxspeedz)
    
    if action.action == ACTION_STARTGAME:
      from py2play.idler import IDLER
      if IDLER.waiting_until_ready: IDLER.start_game()
      else:                         IDLER.unpause()
      
    Character.begin_action(self, action)
    
  def get_controler(self):
    return self.player.controler
    if self.player.active: return self.player.controler
    return None
  controler = property(get_controler)
  
  def hurt(self, life):
    if   self.level.difficulty == 0: return Character.hurt(self, 0.6 * life)
    elif self.level.difficulty == 1: return Character.hurt(self, 0.9 * life)
    else:                            return Character.hurt(self, life)
    
  def die(self):
    if len(explosions) < 4: # Limit the number of simultaneous explosions
      explode = Explosion(self.parent)
      explode.move(self)
      explode.life = 60
      explode.particle_width = explode.particle_height = 0.7
      
      sound.play("explose-3.wav", explode)
      
    oldlevel = self.level
    self.level.remove_character(self)
    self.player.level = oldlevel # May be required !!!
    
    if self.player.active:
      from py2play.idler import IDLER
      IDLER.show_blackbands()
      IDLER.camera.add_traveling(soya.FixTraveling(IDLER.camera, Vector(IDLER.camera, 0.0, 0.0, -1.0) % IDLER))
      IDLER.message(_("__gameover__"), 100000)
      
      # As the character has been removed from the level, the controler
      # does no longer listen for event such as the "escape" key.
      AllowQuit(IDLER)
      
  def __getstate__(self):
    state = Character.__getstate__(self)
    return state[0], self._treat_getstate(state[1])
  
  def __setstate__(self, state):
    Character.__setstate__(self, state)
    self._treat_setstate(state)
    #py2play_character.PlayerCharacter.__setstate__(self, state)
    
    # Loaded => imply this character is a phantom one !!!
    from slune.controler import PhantomControler
    self.player.controler = PhantomControler()
    
    if globdef.VERIFICATION_SERVER: self.logon_verif()
    
  def __repr__(self):
    if self.player: return "<Vehicle %s@%s:%s>" % (self.player.name, self.player.host, self.player.port)
    

explosions = []
class Explosion(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 30):
    soya.Particles.__init__(self, parent, material, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((0.9, 0.0, 0.2, 0.4), (0.9, 0.2, 0.2, 0.8), (0.8, 0.7, 0.1, 0.9), (0.6, 0.4, 0.2, 0.2))
    self.set_sizes((1.0, 1.0))
    self.life = 25
    explosions.append(self)
    
  def generate(self, index):
    angle = random.random() * 3.1417
    sx = math.cos(angle)
    sy = -math.sqrt(random.random() * 3.0)
    sz = math.sin(angle)
    l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, 0.5 + random.random(), sx * l, sy * l, sz * l, 0.0, 0.01, 0.0)

  def begin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.removable = 1
      explosions.remove(self)
      self.life = -1
      
class FlameThrower(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 35):
    soya.Particles.__init__(self, parent, material, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((0.0, 0.8, 0.0, 0.4), (0.1, 0.8, 0.1, 0.9), (0.8, 0.1, 0.1, 0.9), (0.1, 0.1, 0.8, 0.9), (0.2, 0.2, 0.8, 0.2))
    self.set_sizes((2.0, 2.0))
    self.life = 30
    explosions.append(self)
    
  def generate(self, index):
    angle = random.random() * 3.1417
    v = Vector(self,
               random.uniform(-0.3,  0.3),
               random.uniform(-0.3,  0.3),
               random.uniform(-1.5, -1.0),
               )
    v.convert_to(self.parent.parent)
    a = Vector(self, 0.0, 0.00, 0.02)
    a.convert_to(self.parent.parent)
    
    self.set_particle(index, 0.5 + random.random(), v.x, v.y, v.z, a.x, a.y, a.z)

  def begin_round(self):
    soya.Particles.begin_round(self)
    
    if (self.life > 0) and self.parent:
      for character in self.parent.level.pushables:
        cx, cy, cz = self.transform_point(character.x, character.y, character.z, character.parent)
        if (-20.0 < cz < -0.2) and (-0.5 < cy / cz < 0.5) and (-0.5 < cx / cz < 0.5):
          character.hurt(-0.1 / (cz - 1.0))
          _V.set_start_end(self, character)
          _V.convert_to(character.parent)
          _V.set_length(-1.0 / (cz - 1.0) / character.weight)
          character.push(_V.x, _V.y + 0.04, _V.z, self.parent)
          
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      self.removable = 1
      explosions.remove(self)
      self.life = -1
      
      
      
class Nitro(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 50):
    soya.Particles.__init__(self, parent, material, nb_particles)
    self.set_colors((0.4, 0.7, 0.8, 0.4), (0.2, 0.7, 0.9, 0.8), (0.2, 0.9, 0.9, 0.9), (0.1, 0.3, 0.3, 0.2))
    self.set_sizes((1.0, 1.0))
    self.life = 25
    self.max_particles_per_round = 4
    self.auto_generate_particle = 1
    
  def generate(self, index):
    angle = random.random() * 3.1417
    sx = math.cos(angle)
    sy = -math.sqrt(random.random() * 3.0)
    sz = math.sin(angle)
    l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
    #self.set_particle(index, 0.5 + random.random(), 0.0, 0.0, 0.0, random.random() - 0.5, random.random() - 0.5, random.random() - 0.5)
    
    pos = self.get_root().transform_vector((random.random() - 0.5) * 0.2, (random.random() - 0.5) * 0.2, 0.2, self)
    self.set_particle(index, 0.5 + random.random(), pos[0], pos[1], pos[2], 0.0, 0.0, 0.0)
      
# class Explosion(particle.SpriteSystem):
#   def __init__(self, parent = None, material = None, nb_particles = 30):
#     particle.SpriteSystem.__init__(self, parent, particle.FireGenerator(initial_speed = 0.4), material or particle._default(), nb_particles = nb_particles)
#     self.particle_width = self.particle_height = 0.5
#     self.life           = 25
#     explosions.append(self)
    
#   def begin_round(self):
#     if   self.life >  0: self.life -= 1
#     elif self.life == 0:
#       self.generator = None
#       self.removable = 1
#       explosions.remove(self)
#       self.life = -1
      
# class Snow(particle.Particles):
#   def __init__(self, parent = None, material = None, nb_particles = 10):
#     particle.Particles.__init__(self, parent, material or particle._default(), nb_particles)
#     self.auto_generate_particle = 1
#     self.removable = 1
#     self.set_colors((0.6, 0.75, 0.8, 0.2), (0.6, 0.75, 0.8, 0.6), (0.4, 0.7, 0.8, 0.2))
#     self.set_sizes((1.0, 1.0))
#     self.life = 10
    
#   def generate(self, index):
#     angle = random.random() * 3.1417
#     sx = math.cos(angle)
#     sy = -math.sqrt(random.random() * 3.0)
#     sz = math.sin(angle)
#     l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
#     self.set_particle(index, 0.5 + random.random(), sx * l, sy * l, sz * l, 0.0, 0.01, 0.0)
    
#   def begin_round(self):
#     particle.Particles.begin_round(self)
#     if   self.life >  0: self.life -= 1
#     elif self.life == 0:
#       self.auto_generate_particle = 0
#       self.removable = 1
#       self.life = -1
    
class LifeBar(widget.Widget):
  def __init__(self, character = None):
    widget.Widget.__init__(self)
    self.character = character
    
    self.left   = soya.get_screen_width() - 30.0
    self.top    = 10.0
    self.width  = 20.0
    self.height = soya.get_screen_height() - 20.0
    
  def render(self):
    if self.character:
      import soya.opengl as soyaopengl
      soyaopengl.glBegin(soyaopengl.GL_QUADS)
      soyaopengl.glColor4f(0.7, 0.0, 0.0, 1.0)
      
      soyaopengl.glVertex2f(self.left             , self.top + self.height)
      soyaopengl.glVertex2f(self.left + self.width, self.top + self.height)
      
      top = self.top + (1.0 - self.character.life) * self.height
      soyaopengl.glColor4f(1.0, 0.7, 0.0, 1.0)
      soyaopengl.glVertex2f(self.left + self.width, top)
      soyaopengl.glVertex2f(self.left             , top)
      soyaopengl.glEnd()

class BossLifeBar(widget.Widget):
  def __init__(self, character = None):
    widget.Widget.__init__(self)
    self.character = character
    
    self.left   = 20.0
    self.top    = 10.0
    self.width  = 20.0
    self.height = soya.get_screen_height() - 20.0
    
  def render(self):
    if self.character:
      import soya.opengl as soyaopengl
      soyaopengl.glBegin(soyaopengl.GL_QUADS)
      soyaopengl.glColor4f(0.0, 0.5, 0.0, 1.0)
      
      soyaopengl.glVertex2f(self.left             , self.top + self.height)
      soyaopengl.glVertex2f(self.left + self.width, self.top + self.height)
      
      top = self.top + (1.0 - self.character.life) * self.height
      soyaopengl.glColor4f(0.7, 1.0, 0.0, 1.0)
      soyaopengl.glVertex2f(self.left + self.width, top)
      soyaopengl.glVertex2f(self.left             , top)
      soyaopengl.glEnd()


class AllowQuit(soya.CoordSyst):
  def begin_round(self):
    soya.CoordSyst.begin_round(self)
    
    import soya.sdlconst as soyasdlconst
    
    for event in soya.process_event():
      if (event[0] == soyasdlconst.MOUSEBUTTONDOWN) or (event[0] == soyasdlconst.KEYDOWN) or (event[0] == soyasdlconst.JOYBUTTONDOWN):
        from py2play.idler import IDLER
        IDLER.end_game()
        




# Slune
# Copyright (C) 2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya, soya.widget as widget
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
import slune.character as slune_character
import slune.controler


def init_bowling_game(level):
  pass
  
def start_bowling_game(level, add_opponent1 = 1):
  from py2play.idler import IDLER
  
  bowling = Bowling(level, level["ball"], 3 - level.difficulty)
  IDLER.no_blackbands_group.insert(0, BowlingLabel(bowling))
  
def end_bowling_game(level):
  pass


def GoBack(character):
  for i in range(20):
    if character.z >= 10.5: break
    
    p = soya.Point(character.level, 10.0, 0.5, 12.0)
    p.convert_to(character)
    
    if   p.x < -4.0: x =  0.7
    elif p.x >  4.0: x = -0.7
    else:            x =  0.0

    if p.z > 0: y = 2.0
    else:       y = 0.0

    yield slune_character.Action(slune_character.ACTION_WAIT, x, y)
  
class Bowling(soya.Volume):
  def __init__(self, parent, ball, nb_extra_ball):
    soya.Volume.__init__(self, parent)
    
    self.ball          = ball
    self.ball_fired    = 0
    self.nb_extra_ball = nb_extra_ball
    self.nb_quilles    = 0
    self.nb_balls_used = 1
    self.ball_pos      = ball.position()
    self.waiting       = 0
    
  def add_ball(self):
    self.nb_extra_ball -= 1
    self.nb_balls_used += 1

    ball = slune_character.Ball(self.parent)
    ball.name = "ball"
    ball.move(self.ball_pos)
    ball.set_shape(self.ball.shape)
    ball.radius   = self.ball.radius
    ball.radius_y = self.ball.radius_y
    ball.weight   = self.ball.weight
    self.parent.pushables.append(ball)
    self.ball = ball
    
    self.ball_fired = 0
    
  def begin_round(self):
    from py2play.idler import IDLER
    
    level = self.parent
    
    for character in level.characters:
      if character.z < -15.0:
        IDLER.level_completed(character, 0)
        level.remove(self)
    
    self.nb_quilles = len([e for e in level if getattr(e, "name", None) == "quille"])
    
    if self.waiting:
      self.waiting -= 1
      if not self.waiting:
        if level.characters[0].z < 10.0:
          level.characters[0].controler.append(GoBack(level.characters[0]))
          self.waiting = 10
        else:
          self.add_ball()
          
    elif not self.ball_fired:
      if self.ball.worth_playing: self.ball_fired = 1
      
    else:
      for e in level:
        if hasattr(e, "name") and ((e.name == "ball") or (e.name == "quille")):
          if e.worth_playing and e.parent: break
      else: # Current ball is finished
        if self.nb_quilles:
          if self.nb_extra_ball:
            self.waiting = 1
          else:
            IDLER.level_completed(character, 0)
            level.remove(self)
            
        else: # All quille are out !
          if self.nb_balls_used == 1:
            IDLER.level_completed(character, 1, _("__bowling_winner_strike__"))
          else:
            IDLER.level_completed(character, 1)
          level.remove(self)
          
  
class BowlingLabel(widget.Label):
  def __init__(self, bowling):
    self.bowling = bowling
    self.format = str(_("__bowlingformat__"))
    print self.format
    self.old_nb_balls   = -1
    self.old_nb_quilles = -1
    widget.Label.__init__(self)
    
    self.left = 10
    self.width  = 700
    self.height = 100
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.top = parent_top + soya.root_widget.height - 43
    
  def widget_begin_round(self):
    if (self.old_nb_balls != self.bowling.nb_extra_ball) or (self.old_nb_quilles != self.bowling.nb_quilles):
      self.text = self.format % (self.bowling.nb_extra_ball, self.bowling.nb_quilles)
      self.old_nb_balls   = self.bowling.nb_extra_ball
      self.old_nb_quilles = self.bowling.nb_quilles



#---------------
# town generator
#---------------
# written by Bertrand 'BLam' Lamy (blam@tuxfamily.org)
# use this code to make Slune level faster!!!

import math
import random
import Tkinter
import Image

import soya, soya.editor


def distance(p1, p2):
  x = p1[0] - p2[0]
  y = p1[1] - p2[1]
  z = p1[2] - p2[2]
  return math.sqrt(x * x + y * y + z * z)


def intersect_2D_lines(p1, v1, p2, v2):
  k = ((p2[1] - p1[1]) * v2[0] + (p1[0] - p2[0]) * v2[1]) / (v1[1] * v2[0] - v1[0] * v2[1])
  # t = ((p2[1] - p1[1]) * v1[0] + (p1[0] - p2[0]) * v1[1]) / (v2[0] * v1[1] - v2[1] * v1[0])
  return (k * v1[0] + p1[0], k * v1[1] + p1[1])


def distance_point_line(point, line_point, line_vector):
  perpend = (-line_vector[1], line_vector[0])
  p1 = point
  v1 = perpend
  p2 = line_point
  v2 = line_vector
  return ((p2[1] - p1[1]) * v2[0] + (p1[0] - p2[0]) * v2[1]) / (v1[1] * v2[0] - v1[0] * v2[1])


def side_of_line(point_to_test, line_point, line_vector, point_inside):
  d1 = distance_point_line(point_to_test, line_point, line_vector)
  d2 = distance_point_line(point_inside, line_point, line_vector)
  if (d1 > 0 and d2 > 0) or (d1 < 0 and d2 < 0): return 1
  return 0
  

def point_is_inside(point_to_test, points):
  i = 0
  while i < len(points):
    p1 = points[i - 1]
    p2 = points[i]
    j = i + 1
    if j >= len(points): j = 0
    p3 = points[j]
    v = (p2[0] - p1[0], p2[1] - p1[1])
    if side_of_line(point_to_test, p1, v, p3) == 0: return 0
    i += 1
  return 1



def translate_points(points, x, z):
  ps = []
  for p in points:
    ps.append((p[0] + x, p[1] + z))
  return ps
    
def rotate_points(points, angle):
  a = angle / 180.0 * math.pi
  coss = math.cos(a)
  sinn = math.sin(a)
  ps = []
  for p in points:
    ps.append((p[0] * coss - p[1] * sinn, p[1] * coss + p[0] * sinn))
  return ps

def make_square(w, d):
  x = w * 0.5
  z = d * 0.5
  return [(-x, -z), (-x, z), (x, z), (x, -z)]

def make_octogon(x, z):
  p = [(x, z), (z, x), (-z, x), (-x, z), (-x, -z), (-z, -x), (z, -x), (x, -z)]
  p.reverse()
  return p

class House:

  def __init__(self, points = [], height = 0.0):
    self.points = points
    self.height = height
    # Section drawer
    self.section = None
    self.section_size = -1.0
    self.materials = None
    # list materials: [(material, x fill style, x scale, y fill style, y scale), ...]
    #   fill style: 'complete', 'plak'
    self.roof = None
    #   roof: ('pyramid', material, height, x scale, y fill style, y scale)
    #         ('trapez',  material1, material2, height, border_incline, x fill scale, y fill scale)
    #         ('flat',    material)

  def translate_points(self, x, z):
    points = []
    for p in self.points:
      points.append((p[0] + x, p[1] + z))
    self.points = points

  def scale_points(self, f):
    points = []
    for p in self.points:
      points.append((p[0] * f, p[1] * f))
    self.points = points

  def rotate_points(self, angle):
    a = angle / 180.0 * math.pi
    coss = math.cos(a)
    sinn = math.sin(a)
    points = []
    for p in self.points:
      points.append((p[0] * coss - p[1] * sinn, p[1] * coss + p[0] * sinn))
    self.points = points

  def make_square(self, w, d):
    x = w * 0.5
    z = d * 0.5
    self.points = [(-x, -z), (-x, z), (x, z), (x, -z)]

  def draw_sections(self, world):
    # create main sections
    section_points = []
    w = soya.World(world)
    i = 0
    while i < len(self.points):
      # section is (x, z)
      curr_section = self.points[i]
      if i + 1 >= len(self.points):
        next_section = self.points[0]
      else:
        next_section = self.points[i + 1]
      prev_section = self.points[i - 1]
      prev_vector = (curr_section[0] - prev_section[0], curr_section[1] - prev_section[1])
      curr_vector = (next_section[0] - curr_section[0], next_section[1] - curr_section[1])
      list = []
      for point in self.section:
        # point is (x, y)
        if point[0] == 0.0:
          list.append((curr_section[0], self.height + point[1], curr_section[1]))
        else:
          # compute point for current section
          w.set_identity()
          p = soya.Point(w, point[0], point[1], 0.0)
          w.look_at(soya.Point(world, curr_vector[0], 0.0, curr_vector[1]))
          cp = p % world
          # compute point for previous section
          w.set_identity()
          p = soya.Point(w, point[0], point[1], 0.0)
          w.look_at(soya.Point(world, prev_vector[0], 0.0, prev_vector[1]))
          pp = p % world
          # compute intersection
          p = intersect_2D_lines((cp.x, cp.z), curr_vector, (pp.x, pp.z), prev_vector)
          # translate
          list.append((p[0] + curr_section[0], self.height + point[1], p[1] + curr_section[1]))
      section_points.append(list)
      i += 1
    world.remove(w)

    # create sub sections
    if self.section_size <= 0.0:
      sub_sections = []
      i = 0
      while i < len(self.points):
        sub_section = [section_points[i - 1], section_points[i]]
        sub_sections.append(sub_section)
        i += 1
    else:
      sub_sections = []
      i = 0
      while i < len(self.points):
        sub_section = []
        prev_section = section_points[i - 1]
        curr_section = section_points[i]
        # add previous section
        sub_section.append(prev_section)
        # compute intermediary section
        prev = self.points[i - 1]
        curr = self.points[i]
        vector = (curr[0] - prev[0], curr[1] - prev[1])
        perpend = (-vector[1], vector[0])
        size = math.sqrt(vector[0] * vector[0] + vector[1] * vector[1])
        nb = int(round(size / self.section_size)) - 1
        j = 0
        while j < nb:
          section = []
          f = (j + 1.0) / (nb + 1.0)
          k = 0
          while k < len(self.section):
            p = intersect_2D_lines(
              (prev_section[k][0], prev_section[k][2]),
              vector,
              (prev[0] + vector[0] * f, prev[1] + vector[1] * f),
              perpend)
            section.append((p[0], prev_section[k][1], p[1]))
            k += 1
          sub_section.append(section)
          j += 1
        # add current section          
        sub_section.append(curr_section)
        sub_sections.append(sub_section)
        i += 1

    # create faces
    for sections in sub_sections:
      i = 0
      while i < len(sections) - 1:
        curr_section = sections[i]
        next_section = sections[i + 1]
        j = 0
        while j < len(curr_section) - 1:
          p1 = curr_section[j]
          p2 = next_section[j]
          p3 = next_section[j + 1]
          p4 = curr_section[j + 1]
          if self.materials:
            material_data = self.materials[j]
            if material_data[1] == 'complete':
              u1 = 0.0
              u2 = material_data[2]
              u3 = u2
              u4 = u1
            elif material_data[1][:4] == 'plak':
              v = (p2[0] - p1[0], p2[2] - p1[2])
              d = 1.0 / math.sqrt(v[0] * v[0] + v[1] * v[1])
              vector = (v[0] * d, v[1] * d)
              perpend = (-vector[1], vector[0])
              ref = intersect_2D_lines((p1[0], p1[2]), vector,
                                       (0.0, 0.0), perpend)
              if vector[0] == 0.0:
                u1 = (p1[2] - ref[1]) / vector[1]
                u2 = (p2[2] - ref[1]) / vector[1]
              else:
                u1 = (p1[0] - ref[0]) / vector[0]
                u2 = (p2[0] - ref[0]) / vector[0]
              ref = intersect_2D_lines((p4[0], p4[2]), vector,
                                       (0.0, 0.0), perpend)
              if vector[0] == 0.0:
                u3 = (p3[2] - ref[1]) / vector[1]
                u4 = (p4[2] - ref[1]) / vector[1]
              else:
                u3 = (p3[0] - ref[0]) / vector[0]
                u4 = (p4[0] - ref[0]) / vector[0]
              u1 *= material_data[2]
              u2 *= material_data[2]
              u3 *= material_data[2]
              u4 *= material_data[2]
            if material_data[1] == 'plak round':
              u1 = round(u1)
              u2 = round(u2)
              u3 = round(u3)
              u4 = round(u4)
            if material_data[3] == 'complete':
              v1 = material_data[4]
              v2 = 0.0
            elif material_data[3] == 'plak':
              v1 = -p1[1]
              v2 = -p3[1]
              v1 *= material_data[4]
              v2 *= material_data[4]
            f = soya.Face(
              world,
              [soya.Vertex(world, p1[0], p1[1], p1[2], u1, v1),
               soya.Vertex(world, p2[0], p2[1], p2[2], u2, v1),
               soya.Vertex(world, p3[0], p3[1], p3[2], u3, v2),
               soya.Vertex(world, p4[0], p4[1], p4[2], u4, v2)], material_data[0])
          else:
            f = soya.Face(
              world,
              [soya.Vertex(world, p1[0], p1[1], p1[2]),
               soya.Vertex(world, p2[0], p2[1], p2[2]),
               soya.Vertex(world, p3[0], p3[1], p3[2]),
               soya.Vertex(world, p4[0], p4[1], p4[2])])
          j += 1
        i += 1

    # create roof
    if self.roof:

      if self.roof[0] == 'pyramid':
        x = 0.0
        z = 0.0
        for point in self.points:
          x += point[0]
          z += point[1]
        center = (x / len(self.points), self.section[-1][1] + self.roof[2], z / len(self.points))
        for sections in sub_sections:
          i = 0
          while i < len(sections) - 1:
            p1 = sections[i][-1]
            p2 = sections[i + 1][-1]
            if self.roof[1]:
              if self.roof[4] == 'plak':
                v1 = distance(p1, center)
                v2 = distance(p2, center)
              elif self.roof[4] == 'complete':
                v1 = self.roof[5]
                v2 = self.roof[5]
              v = (p2[0] - p1[0], p2[2] - p1[2])
              d = 1.0 / math.sqrt(v[0] * v[0] + v[1] * v[1])
              vector = (v[0] * d, v[1] * d)
              perpend = (-vector[1], vector[0])
              ref = intersect_2D_lines((sections[0][-1][0], sections[0][-1][2]), vector, (center[0], center[2]), perpend)
              if vector[0] == 0.0:
                u1 = (p1[2] - ref[1]) / vector[1]
                u2 = (p2[2] - ref[1]) / vector[1]
              else:
                u1 = (p1[0] - ref[0]) / vector[0]
                u2 = (p2[0] - ref[0]) / vector[0]
              u1 *= self.roof[3]
              u2 *= self.roof[3]
              f = soya.Face(
                world,
                [soya.Vertex(world, p1[0], p1[1], p1[2], u1, v1),
                 soya.Vertex(world, p2[0], p2[1], p2[2], u2, v2),
                 soya.Vertex(world, center[0], center[1], center[2], 0.0, 0.0)
                 ], self.roof[1])
            else:
              pass
              f = soya.Face(
                world,
                [soya.Vertex(world, p1[0], p1[1], p1[2]),
                 soya.Vertex(world, p2[0], p2[1], p2[2]),
                 soya.Vertex(world, center[0], center[1], center[2])])
            i += 1

      elif self.roof[0] == 'flat':
        # simple implementation: self.points must be convex or given in the good order
        points = []
        for sections in sub_sections:
          points.append(sections[0][-1])
        points.append(sub_sections[0][0][-1])
        i = 0
        while i + 3 <= len(points):
          p1 = points[i]
          p2 = points[i + 1]
          p3 = points[i + 2]
          if i + 4 <= len(points):
            # make a quad
            p4 = points[i + 3]
            f = soya.Face(
              world,
              [soya.Vertex(world, p1[0], p1[1], p1[2], p1[0], p1[2]),
               soya.Vertex(world, p2[0], p2[1], p2[2], p2[0], p2[2]),
               soya.Vertex(world, p3[0], p3[1], p3[2], p3[0], p3[2]),
               soya.Vertex(world, p4[0], p4[1], p4[2], p4[0], p4[2])],
               self.roof[1])
            i += 3
          else:
            # make a triangle
            f = soya.Face(
              world,
              [soya.Vertex(world, p1[0], p1[1], p1[2], p1[0], p1[2]),
               soya.Vertex(world, p2[0], p2[1], p2[2], p2[0], p2[2]),
               soya.Vertex(world, p3[0], p3[1], p3[2], p3[0], p3[2])],
               self.roof[1])
            i += 2
        
      elif self.roof[0] == 'trapez':
        spine_points = []
        base_points  = []
        for sections in sub_sections:
          base_points.append(sections[0][-1])
        if len(base_points) == 4:
          
          d1 = distance(base_points[0], base_points[1])
          d2 = distance(base_points[0], base_points[-1])
          if d2 < d1:
            p = base_points[-1]
            base_points.remove(p)
            base_points.insert(0, p)
          spine_points = (
            ((base_points[0][0] + base_points[1][0]) * 0.5,
             (base_points[0][1] + base_points[1][1]) * 0.5 + self.roof[3],
              (base_points[0][2] + base_points[1][2]) * 0.5),
            ((base_points[2][0] + base_points[3][0]) * 0.5,
             (base_points[2][1] + base_points[3][1]) * 0.5 + self.roof[3],
             (base_points[2][2] + base_points[3][2]) * 0.5)
            )
          ref = spine_points[0]
          # incline border
          v = (spine_points[1][0] - spine_points[0][0],
               spine_points[1][1] - spine_points[0][1],
               spine_points[1][2] - spine_points[0][2])
          d = 1.0 / math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
          v = (v[0] * d, v[1] * d, v[2] * d)
          f = self.roof[4]
          spine_points = (
            (spine_points[0][0] + f * v[0], spine_points[0][1] + f * v[1], spine_points[0][2] + f * v[2]),
            (spine_points[1][0] - f * v[0], spine_points[1][1] - f * v[1], spine_points[1][2] - f * v[2]))
          # draw 2 triangles
          u = 0.5 * distance(base_points[0], base_points[1]) * self.roof[5]
          v1 = soya.Vertex(world, spine_points[0][0], spine_points[0][1], spine_points[0][2], 0.0, 0.0)
          v2 = soya.Vertex(world, base_points[0][0], base_points[0][1], base_points[0][2],   -u, self.roof[6])
          v3 = soya.Vertex(world, base_points[1][0], base_points[1][1], base_points[1][2],    u, self.roof[6])
          f = soya.Face(world, [v1, v2, v3], self.roof[2])
          v1 = soya.Vertex(world, spine_points[1][0], spine_points[1][1], spine_points[1][2], 0.0, 0.0)
          v2 = soya.Vertex(world, base_points[2][0], base_points[2][1], base_points[2][2],   -u, self.roof[6])
          v3 = soya.Vertex(world, base_points[3][0], base_points[3][1], base_points[3][2],    u, self.roof[6])
          f = soya.Face(world, [v1, v2, v3], self.roof[2])

          # draw 2 quads

          u = distance(spine_points[0], ref) * self.roof[5]
          v1 = soya.Vertex(world, spine_points[0][0], spine_points[0][1], spine_points[0][2], u, 0.0)
          v2 = soya.Vertex(world, base_points[1][0], base_points[1][1], base_points[1][2],  0.0, self.roof[6])
          u = distance(base_points[2], base_points[1]) * self.roof[5]
          v3 = soya.Vertex(world, base_points[2][0], base_points[2][1], base_points[2][2],    u, self.roof[6])
          u = distance(spine_points[1], ref) * self.roof[5]
          v4 = soya.Vertex(world, spine_points[1][0], spine_points[1][1], spine_points[1][2], u, 0.0)
          f = soya.Face(world, [v1, v2, v3, v4], self.roof[1])

          u = distance(spine_points[1], ref) * self.roof[5]
          v1 = soya.Vertex(world, spine_points[1][0], spine_points[1][1], spine_points[1][2], u, 0.0)
          u = distance(base_points[0], base_points[3]) * self.roof[5]
          v2 = soya.Vertex(world, base_points[3][0], base_points[3][1], base_points[3][2],    u, self.roof[6])
          v3 = soya.Vertex(world, base_points[0][0], base_points[0][1], base_points[0][2],  0.0, self.roof[6])
          u = distance(spine_points[0], ref) * self.roof[5]
          v4 = soya.Vertex(world, spine_points[0][0], spine_points[0][1], spine_points[0][2], u, 0.0)
          f = soya.Face(world, [v1, v2, v3, v4], self.roof[1])


  def draw(self, world):
    if self.section:
      self.draw_sections(world)

  def set_on_land (self, land, house_world = None):
    factor = land.scale_factor
    x1 = self.points[0][0]
    z1 = self.points[0][1]
    x2 = x1
    z2 = z1
    for p in self.points:
      if p[0] < x1: x1 = p[0]
      if p[0] > x2: x2 = p[0]
      if p[1] < z1: z1 = p[1]
      if p[1] > z2: z2 = p[1]
    x1 = math.floor(x1 / factor)
    z1 = math.floor(z1 / factor)
    x2 = math.ceil(x2 / factor)
    z2 = math.ceil(z2 / factor)
    j = z1
    min_h = 10000000.0
    while j < z2:
      i = x1
      while i < x2:
        if point_is_inside((i * factor, j * factor), self.points) == 1:
          land.add_point_option(int(i), int(j), 1)
          h = land.get_height(int(i), int(j))
          if h < min_h: min_h = h
        i += 1
      j += 1
    if min_h == 10000000.0:
      min_h = 0.0
# hack
#    if house_world: house_world.y = min_h


    
def make_city(filename_buildings_map, filename_styles_map, styles_dict, land = None, scale = 1.0):

  buildings = soya.World()

  def make_building(points, style = None):
    v = soya.World(buildings)
    h = House(points)
    h.scale_points(scale)

    if style:
      style_func = styles_dict[style]
    else:
      style_func = styles_dict[int(random.random() * len(styles_dict))]
    style_func(h)

    h.draw(v)
    if land: h.set_on_land(land, v)



  # auto make buildings from image
  #   RED  building has only vertical or horizontal walls
  #   BLUE building uses the extreme points in the 2 dimensions of the filled surface

  img = Image.open(filename_buildings_map)
  map = []
  i = 0
  while(i < img.size[0]):
    j = 0
    col = []
    while(j < img.size[1]):
      if img.getpixel((i, j)) == (255, 0, 0):
        col.append(1)
      elif img.getpixel((i, j)) == (0, 0, 255):
        col.append(2)
      else:
        col.append(0)
      j += 1
    map.append(col)
    i += 1

  if filename_styles_map: img_styles = Image.open(filename_styles_map)

  def map_fill(map, x, y):
    if map[x][y] == 1:
      map[x][y] = 0
      map_fill(map, x - 1, y)
      map_fill(map, x + 1, y)
      map_fill(map, x, y - 1)
      map_fill(map, x, y + 1)

  def red_house(map, X, Y):
    global x, y
    global startx, starty
    x = X
    y = Y
    startx = X
    starty = Y
    points = [(float(x), float(y))]
    def search_left():
      global x, y, startx, starty
      while 1:
        if x == startx and y == starty: return
        x -= 1
        if map[x][y] == 0:
          x += 1
          points.append((float(x), float(y)))
          search_up()
          return
        if map[x][y + 1] == 1:
          points.append((float(x), float(y)))
          search_down()
          return
    def search_right():
      global x, y, startx, starty
      while 1:
        if x == startx and y == starty: return
        x += 1
        if map[x][y] == 0:
          x -= 1
          points.append((float(x), float(y)))
          search_down()
          return
        if map[x][y - 1] == 1:
          points.append((float(x), float(y)))
          search_up()
          return
    def search_up():
      global x, y, startx, starty
      while 1:
        if x == startx and y == starty: return
        y -= 1
        if map[x][y] == 0:
          y += 1
          points.append((float(x), float(y)))
          search_right()
          return
        if map[x - 1][y] == 1:
          points.append((float(x), float(y)))
          search_left()
          return
    def search_down():
      global x, y, startx, starty
      while 1:
        if x == startx and y == starty: return
        y += 1
        if map[x][y] == 0:
          y -= 1
          points.append((float(x), float(y)))
          search_left()
          return
        if map[x + 1][y] == 1:
          points.append((float(x), float(y)))
          search_right()
          return
    x += 1
    search_right()
    points.reverse()
    map_fill(map, X, Y)
    return points
            
  def blue_house(map, X, Y):
    points = [(X, Y), (X, Y), (X, Y), (X, Y)]
    def search_extremes(map, x, y, list):
      if map[x][y] == 2:
        minx = list[0][0]
        miny = list[1][1]
        maxx = list[2][0]
        maxy = list[3][1]
        if x == minx:
          if len(list[0]) == 3:
            y1 = list[0][1]
            y2 = list[0][2]
            if y < y1: list[0] = (minx, y, y2)
            if y > y2: list[0] = (minx, y1, y)
          else:
            y0 = list[0][1]
            if y < y0: list[0] = (minx, y, y0)
            if y > y0: list[0] = (minx, y0, y)
        if y == miny:
          if len(list[1]) == 3:
            x1 = list[1][0]
            x2 = list[1][2]
            if x < x1: list[1] = (x, miny, x2)
            if x > x2: list[1] = (x1, miny, x)
          else:
            x0 = list[1][0]
            if x < x0: list[1] = (x, miny, x0)
            if x > x0: list[1] = (x0, miny, x)
        if x == maxx:
          if len(list[2]) == 3:
            y1 = list[2][1]
            y2 = list[2][2]
            if y < y1: list[2] = (maxx, y, y2)
            if y > y2: list[2] = (maxx, y1, y)
          else:
            y0 = list[2][1]
            if y < y0: list[2] = (maxx, y, y0)
            if y > y0: list[2] = (maxx, y0, y)
        if y == maxy:
          if len(list[3]) == 3:
            x1 = list[3][0]
            x2 = list[3][2]
            if x < x1: list[3] = (x, maxy, x2)
            if x > x2: list[3] = (x1, maxy, x)
          else:
            x0 = list[3][0]
            if x < x0: list[3] = (x, maxy, x0)
            if x > x0: list[3] = (x0, maxy, x)
        if x < list[0][0]: list[0] = (x, y)
        if x > list[2][0]: list[2] = (x, y)
        if y < list[1][1]: list[1] = (x, y)
        if y > list[3][1]: list[3] = (x, y)
        map[x][y] = 0
        search_extremes(map, x - 1, y, list)
        search_extremes(map, x + 1, y, list)
        search_extremes(map, x, y - 1, list)
        search_extremes(map, x, y + 1, list)
        
    search_extremes(map, X, Y, points)

    pts = []
    p = points[0]
    if len(p) == 3:
      pts.append((float(p[0]), float(p[2])))
    pts.append((float(p[0]), float(p[1])))
    p = points[1]
    pts.append((float(p[0]), float(p[1])))
    if len(p) == 3:
      pts.append((float(p[2]), float(p[1])))
    p = points[2]
    pts.append((float(p[0]), float(p[1])))
    if len(p) == 3:
      pts.append((float(p[0]), float(p[2])))
    p = points[3]
    if len(p) == 3:
      pts.append((float(p[2]), float(p[1])))
    pts.append((float(p[0]), float(p[1])))

    points = []
    for p in pts:
      if len(points) == 0 or p != points[-1]:
        points.append(p)
    if points[0] == points[-1]: points.remove(points[-1])
            
    points.reverse()
    return points
        

  j = 0
  while(j < img.size[1]):
    i = 0
    while(i < img.size[0]):
      points = None
      if map[i][j] == 1:
        points = red_house(map, i, j)
      elif map[i][j] == 2:
        points = blue_house(map, i, j)
      if points:
        if filename_styles_map:
          styl = img_styles.getpixel((i, j))
        else:
          styl = None
        make_building(points, styl)
      i = i + 1
    j = j + 1
        

  return buildings





# HACK
soya.Material.PATH = "/home/blam/prog/slune/materials"


scene = soya.World()


m1 = soya.Material.get('house3')
m2 = soya.Material.get('inde_2')
m3 = soya.Material.get('bark1')
m4 = soya.Material.get('house3b')
m5 = soya.Material.get('inde_4')
m6 = soya.Material.get('inde_1')
m7 = soya.Material.get('inde_3')
m9 = soya.Material.get('taj3')

ms1 = (m1, 'plak round', 0.5, 'complete', 1.0)
ms2 = (m2, 'plak', 1.0, 'plak', 1.0)
ms3 = (m3, 'plak', 1.0, 'plak', 1.0)
ms4 = (m4, 'plak round', 1.0, 'complete', 1.0)
ms6 = (m6, 'plak round', 1.0, 'complete', 1.0)
ms7 = (m7, 'plak round', 1.0, 'complete', 1.0)
ms8 = (m9, 'plak', 1.0, 'complete', 2.0)
ms9 = (m9, 'plak', 1.0, 'complete', 1.0)



def set_style_1 (house):
  house.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 3.0), (0.2, 3.2), (0.2, 4.2), (0.4, 4.3)]
  house.materials = [ ms2, ms1, ms3, ms4, ms3 ]
  house.roof = ('trapez',  m5, m5, 1.5, 2.0, 0.5, 1.0)
  
def set_style_2 (house):
  house.section = [(0.2, 0.0), (0.2, 1.0), (0.2, 2.0), (0.0, 2.2), (0.0, 3.2), (0.0, 4.2)]
  house.materials = [ ms2, ms6, ms3, ms7, ms2 ]
  house.roof = ('trapez',  m3, m2, 1.5, 0.0, 1.0, 1.5)

def set_style_round (house):
  house.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 3.0), (0.3, 3.3), (0.3, 4.3)]
  house.materials = [ ms6, ms2, ms3, ms7 ]
  house.roof = ('pyramid',  m5, 2.0, 0.5, 'complete', 3.0)


#house = House([(-4.0, -2.0), (-4.0, 2.0), (4.0, 2.0), (4.0, -2.0)])
#house = House([(2.0, 2.0), (2.0, -3.0), (0.0, -3.0), (0.0, 0.0), (-4.0, 0.0), (-4.0, 2.0)])
#house.section_size = -1.0

#house.roof = ('pyramid', soya.Material.get('inde_4'), 2.0, 'complete', 2.0)
#house.roof = ('flat', soya.Material.get('ground1'))
#house.roof = ('trapez',  m5, m5, 1.5, 2.0, 0.5, 1.0)

#set_style_2(house)




# octogon house
#house = House([(-2.0, 5.0), (2.0, 5.0), (5.0, 2.0), (5.0, -2.0), (2.0, -5.0), (-2.0, -5.0), (-5.0, -2.0), (-5.0, 2.0)])

# square house + triangle side
#house = House([(-4.0, -3.0), (-4.0, -2.0), (-5.5, 0.0), (-4.0, 2.0), (-4.0, 3.0), (4.0, 3.0), (4.0, 2.0), (5.5, 0.0), (4.0, -2.0), (4.0, -3.0)])
#house.section = [(0.2, 0.0), (0.2, 2.0), (0.0, 2.2), (0.0, 3.2)]
#house.materials = [ ms8, ms3, ms9 ]
#house.roof = ('pyramid',  m5, 2.0, 'complete', 2.0)
#house.draw(scene)

#scene.filename = "cyberbombay-shape"
#scene.save()

#soya.editor.edit(scene)
#Tkinter.mainloop()




# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math, slune.sound as sound

import soya, soya.widget as widget, soya.ray as ray
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
from soya import Point, Vector
import slune.character as slune_character
import slune.controler

def init_freestyle_game(level):
  pass
      
def start_freestyle_game(level, time_limit = -1.0, no_time_left_message = None):
  competitors = []
  for character in level.characters:
    if isinstance(character, slune_character.Competitor):
      character.score       = 0
      character.score_start = time.time()
      competitors.append(character)
      
  from py2play.idler import IDLER
  IDLER.no_blackbands_group.add(ScoreLabel(competitors, time_limit, no_time_left_message))
  
def end_freestyle_game(level):
  for widget in soya.root_widget.children:
    if isinstance(widget, ScoreLabel):
      soya.root_widget.remove(widget)
      break
    

def start_time_limit(level, time_limits, no_time_left_message = None, start_checker = 0):
  competitors = []
  for character in level.characters:
    if isinstance(character, slune_character.Competitor):
      character.score_start = time.time()
      competitors.append(character)
      
  from py2play.idler import IDLER
  IDLER.no_blackbands_group.add(TimerLabel(competitors, time_limits[level.difficulty], no_time_left_message))
  
  if start_checker:
    Checker(level)
    
    if not hasattr(level, "flags"):
      import race
      
      level.flags = [] # Indexes all flags
      
      for item in level:
        if isinstance(item, race.Flag):
          item.visible = (item.name == "flag_0")
          level.flags.append(item)
          
      level.flags.sort(lambda f1, f2: cmp(f1.flag_id, f2.flag_id))
      
    
def end_time_limit(level):
  for widget in soya.root_widget.children:
    if isinstance(widget, TimerLabel):
      soya.root_widget.remove(widget)
      break
    
    
class Checker(soya.Volume):
  def begin_round(self):
    from py2play.idler import IDLER
    
    level = self.parent
    
    for character in level.characters:
      if isinstance(character, slune_character.Competitor):
        flag = level.flags[character.next_flag_id]
        if character.distance_to(flag) < (flag.radius + character.radius):
          if character.next_flag_id + 1 < len(self.parent.flags):
            next_flag_id = character.next_flag_id + 1
            next_flag = self.parent.flags[next_flag_id]
          else:
            next_flag_id = 0
            next_flag = self.parent.flags[0]
            character.nb_laps += 1
            now = time.time()
            this_lap = now - character.lap_start
            if this_lap < character.best_lap: character.best_lap = this_lap
            character.lap_start = now

            if character.nb_laps >= 0:
              firework = soya.FlagFirework(level, nb_particles = 6)
              firework.move(flag)
              firework.regenerate()
              firework.removable = 1
              
              character.next_flag_id = 0
              
              level.ranks.append(character)

              if len(level.ranks) == 1: # Winner
                IDLER.level_completed(character, 1, _("__race_winner__"))
              else:
                IDLER.level_completed(character, 0, _("__race_looser__") % len(level.ranks))
                
              from py2play.player import CURRENT_PLAYER
              if not character is CURRENT_PLAYER.character:
                IDLER.message(_("__race_arrival__") % (character.displayname(), len(level.ranks)))
                
              if not character.played:
                character.level.remove_character(character)
              return
            
          character.next_flag_id = next_flag_id
          
          sound.play("flag-1.wav", character)
          
          if character.played and character.player.active:
            flag.visible = 0
            next_flag.visible = 1
            
            firework = soya.Smoke(level, nb_particles = 10, removable = 1)
            firework.move(flag)
            firework.regenerate()



V = Vector()

SCORES = {
  "jump"               :    5,
  "longjump"           :   10,
  "verylongjump"       :  150,
  "climbing"           :   10,
  "highjump"           :   15,
  
  "180"                :   50,
  "360"                :   85,
  "720"                :  110,
  "rotup"              :   50,
  "rotup180"           :  120,
  "rotup360"           :  150,
  "rotup720"           : 1000,
  
  "looping"            :  100,
  "doublelooping"      : 1000,
  "retrolooping"       :  200,
  "doubleretrolooping" : 1500,
  
  "s"                  :  100,
  "doubles"            : 1000,
  
  "spinfall"           :  150,
  "bigspinfall"        : 1200,
  "flamingtorch"       :   80,
  "bigflamingtorch"    :  800,
  }

class FreeStyle(soya.Volume):
  def __init__(self, parent = None, character = None):
    soya.Volume.__init__(self, parent)
    self.character = character
    
    self.last_figure        = ""
    self.last_figure_nb     = 0
    self.last_figure_time   = 0
    self.jump_direction     = Vector()
    self.last_180_direction = Vector()
    self.last_180_y         = 1.0
    self.on_ground          = 1
    self.y_angle            = 0.0
    self.y_direction        = Vector()
    self.jump_time          = 0
    self.nb_180             = 0
    self.nb_s               = 0
    self.next_s_dir         = -1.0
    self.jump_y             = 0.0
    self.max_y              = 0.0
    self.spinfall           = 0
    self.flamingtorch       = 0
    
  def recognized(self, figure):
    from py2play.idler import IDLER
    if (self.last_figure == figure) and (self.character.round - self.last_figure_time < 500):
      self.last_figure      = figure
      self.last_figure_nb  += 1
      self.last_figure_time = self.character.round
      
      bonus = self.last_figure_nb * SCORES[figure]
      IDLER.message(_("__figure_" + figure + "__") + " X " + str(self.last_figure_nb) + " (" + str(bonus) + ")")
      
    else:
      self.last_figure      = figure
      self.last_figure_nb   = 1
      self.last_figure_time = self.character.round
      
      bonus = SCORES[figure]
      IDLER.message(_("__figure_" + figure + "__") + " (" + str(bonus) + ")")
      
    self.character.score += bonus
    
  def begin_round(self):
    if self.on_ground and not self.character.on_ground: # Start jumping
      self.jump_time    = 0
      self.on_ground    = 0
      self.y_angle      = 0.0
      self.nb_180       = 0
      self.nb_s         = 0
      self.next_s_dir   = -1.0
      self.spinfall     = 0
      self.flamingtorch = 0
      self.jump_y       = self.character.y
      self.max_y        = self.jump_y
      self.last_180_y   = 1.0
      self.jump_direction.set_xyz(*self.character.level.transform_vector(0.0, 0.0, -1.0, self.character))
      self.last_180_direction.clone(self.jump_direction)
      self.y_direction.clone(self.jump_direction)
      
    elif self.character.on_ground and not self.on_ground: # End jumping
      self.on_ground = 1
      
      if self.jump_time <  20: return # NOT a valid jump !!!
      
      if (self.y_angle >  560.0): self.recognized("doublelooping"); return
      if (self.y_angle >  200.0): self.recognized("looping"); return
      if (self.y_angle < -560.0): self.recognized("doubleretrolooping"); return
      if (self.y_angle < -200.0): self.recognized("retrolooping"); return
      
      if self.nb_s == 2: self.recognized("s"); return
      if self.nb_s == 4: self.recognized("doubles"); return
      
      if self.spinfall > 27: self.recognized("bigspinfall"); return
      if self.spinfall > 10: self.recognized("spinfall"); return
      
      if self.flamingtorch > 30: self.recognized("bigflamingtorch"); return
      if self.flamingtorch > 10: self.recognized("flamingtorch"); return
      
      if   self.y_direction.y > 0.75:
        if self.nb_180 == 1: self.recognized("180"); return
        if self.nb_180 == 2: self.recognized("360"); return
        if self.nb_180 == 4: self.recognized("720"); return
        
      elif self.y_direction.y < -0.75:
        if self.nb_180 == 2: self.recognized("rotup180"); return
        if self.nb_180 == 3: self.recognized("rotup360"); return
        if self.nb_180 == 5: self.recognized("rotup720"); return
        
      if self.y_direction.y < -0.5: self.recognized("rotup"); return

      if self.jump_time > 200:                 self.recognized("verylongjump"); return
      if self.character.y - self.jump_y > 3.0: self.recognized("climbing"); return
      if self.max_y - self.jump_y >  7.0:      self.recognized("highjump"); return
      if self.jump_time >  70:                 self.recognized("longjump"); return
      if self.jump_time >  20:                 self.recognized("jump"); return
      
    if not self.character.on_ground: # Jumping
      self.jump_time += 1
      if self.max_y < self.character.y: self.max_y = self.character.y
      
      x, y, z = self.character.transform_vector(self.y_direction.x, self.y_direction.y, self.y_direction.z, self.character.level)
      
      #self.y_angle -= z
      
      if   self.character.current_action.action == slune_character.ACTION_ROTUP:
        self.y_angle += self.character.rotup_speed
      elif self.character.current_action.action == slune_character.ACTION_ROTDOWN:
        self.y_angle -= self.character.rotup_speed
      
      self.y_direction.set_xyz(*self.character.level.transform_vector(0.0, 1.0,  0.0, self.character))
      if self.y_direction.y * self.next_s_dir > 0.8:
        self.nb_s   += 1
        self.next_s_dir *= -1.0
        
      V.set_xyz(*self.character.level.transform_vector(0.0, 0.0, -1.0, self.character))
      if V.dot_product(self.last_180_direction) < -0.5:
        if self.last_180_y * self.y_direction.y > 0.3:
          self.nb_180 += 1
          self.last_180_direction *= -1
        else:
          self.last_180_y *= -1.0
          
      if self.character.current_action.action == slune_character.ACTION_ROTLAT:
        if   V.y < -0.6: self.spinfall += 1
        elif (V.y >  0.8) and self.character.current_action.mouse_y > 0.8: self.flamingtorch += 1
        

class ScoreLabel(widget.Label):
  def __init__(self, characters, time_limit = -1.0, no_time_left_message = None):
    self.characters = characters
    self.format1 = str(_("__scoreformat1__")) + "\n"
    format2 = str(_("__scoreformat2__"))
    self.format2s = "\n".join([character.displayname() + format2 for character in characters])
    widget.Label.__init__(self)
    
    self.left = 10
    self.top = soya.root_widget.height - 33 * (len(self.characters) + 1)
    self.width  = 400
    self.height = 100
    
    self.time_limit = time_limit
    self.no_time_left_message = no_time_left_message or _("__no_time_left__")
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.top = soya.root_widget.height - 33 * (len(self.characters) + 1)
    
  def render(self):
    if self.time_limit == -1.0:
      time_left = 0.0
    else:
      time_left = (self.time_limit - (time.time() - self.characters[0].score_start))
      if time_left < 0.0:
        self.time_limit = -1.0
        
        import videosequence
        for character in self.characters:
          videosequence.MissionCompleted(0, self.no_time_left_message).start(tux = character)
          
    if len(self.characters) == 1:
      self.text = self.format1 % time_left + self.format2s % self.characters[0].score
    else:
      self.text = self.format1 % time_left + self.format2s % [str(character.score) for character in self.characters]
      
    widget.Label.render(self)
    

class TimerLabel(widget.Label):
  def __init__(self, characters, time_limit = -1.0, no_time_left = None):
    self.characters = characters
    self.format1 = str(_("__scoreformat1__")) + "\n"
    widget.Label.__init__(self)
    
    self.left = 10
    self.top = soya.root_widget.height - 33
    self.width  = 400
    self.height = 100
    
    self.time_limit = time_limit
    self.no_time_left = no_time_left or _("__no_time_left__")
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.top = soya.root_widget.height - 33
    
  def render(self):
    if self.time_limit == -1.0:
      time_left = 0.0
    else:
      time_left = (self.time_limit - (time.time() - self.characters[0].score_start))
      if time_left < 0.0:
        self.time_limit = -1.0

        if callable(self.no_time_left): self.no_time_left(self.characters)
        else: 
          import videosequence
          for character in self.characters:
            videosequence.MissionCompleted(0, self.no_time_left).start(tux = character)
            
    self.text = self.format1 % time_left
      
    widget.Label.render(self)
    
